from __future__ import annotations
import inspect
import types
import typing
from PyQt5.QtCore import QDataStream


class OpenTeleException(BaseException):  # nocov
    """
    Base exception of the library.
    """

    def __init__(self, message: str = None, stack_index: int = 1) -> None:

        super().__init__(message if (message != None) else "")

        self.message = message
        self.desc = self.__class__.__name__

        self.frame = inspect.currentframe()
        for i in range(stack_index):
            self.frame = self.frame.f_back

        self._caller_class = (
            self.frame.f_locals["self"].__class__
            if "self" in self.frame.f_locals
            else None
        )
        self._caller_method = self.frame.f_code.co_name

        if self._caller_method != "<module>":
            args, _, _, locals = inspect.getargvalues(self.frame)
            parameters = {arg: locals[arg] for arg in args}
            self._caller_method_params = "".join(
                f"{i}={parameters[i]}, " for i in parameters
            )[:-2]
        else:
            self._caller_method = "__main__"
            self._caller_method_params = ""

        if self.desc == "OpenTeleException":
            self.desc = "Unexpected Exception"

    def __str__(self):
        reason = self.desc.__str__()

        if self.message != None:
            reason += f": {self.message}"

        reason += " [ Called by "
        if self._caller_class != None:

            parent_list = []
            base = self._caller_class
            while hasattr(base, "__name__"):
                parent_list.append(base.__name__)
                base = base.__base__

            parent_list.reverse()
            reason += "".join(f"{i}." for i in parent_list[1:])
            reason += self._caller_method + "() ]"
            # reason += f"\n>\t{self._caller_method}({self._caller_method_params})"
        else:
            reason += f"{self._caller_method}() ]"
            # reason += f"{self._caller_method}({self._caller_method_params}) ]"
        return reason


class TFileNotFound(OpenTeleException):
    """
    Could not find or open the file
    """


class TDataInvalidMagic(OpenTeleException):
    """
    TData file has an invalid magic data, which is the first 4 bytes of the file\n
    This usually mean that the file is corrupted or not in the supported formats
    """


class TDataInvalidCheckSum(OpenTeleException):
    """
    TData file has an invalid checksum\n
    This usually mean that the file is corrupted or not in the supported formats
    """


class TDataBadDecryptKey(OpenTeleException):
    """
    Could not decrypt the data with this key\n
    This usually mean that the file is password-encrypted
    """


class TDataWrongPasscode(OpenTeleException):
    """
    Wrong passcode to decrypt tdata folder\n
    """


class TDataBadEncryptedDataSize(OpenTeleException):
    """
    The encrypted data size part of the file is corrupted
    """


class TDataBadDecryptedDataSize(OpenTeleException):
    """
    The decrypted data size part of the file is corrupted
    """


class TDataBadConfigData(OpenTeleException):
    """
    TData contains bad config data that couldn't be parsed
    """


class QDataStreamFailed(OpenTeleException):
    """
    Could not stream data from QDataStream\n
    Please check the QDataStream.status() for more information
    """


class AccountAuthKeyNotFound(OpenTeleException):
    """
    Account.authKey is missing, something went wrong
    """


class TDataReadMapDataFailed(OpenTeleException):
    """
    Could not read map data
    """


class TDataReadMapDataIncorrectPasscode(OpenTeleException):
    """
    Could not read map data because of incorrect passcode
    """


class TDataAuthKeyNotFound(OpenTeleException):
    """
    Could not find authKey in TData
    """


class MaxAccountLimit(OpenTeleException):
    """
    Maxed out limit for accounts per tdesktop client
    """


class TDesktopUnauthorized(OpenTeleException):
    """
    TDesktop client is unauthorized
    """


class TelethonUnauthorized(OpenTeleException):
    """
    Telethon client is unauthorized
    """


class TDataSaveFailed(OpenTeleException):
    """
    Could not save TDesktop to tdata folder
    """


class TDesktopNotLoaded(OpenTeleException):
    """
    TDesktop instance has no account
    """


class TDesktopHasNoAccount(OpenTeleException):
    """
    TDesktop instance has no account
    """


class TDAccountNotLoaded(OpenTeleException):
    """
    TDesktop account hasn't been loaded yet
    """


class NoPasswordProvided(OpenTeleException):
    """
    You can't live without a password bro
    """


class PasswordIncorrect(OpenTeleException):
    """
    incorrect passwrd
    """


class LoginFlagInvalid(OpenTeleException):
    """
    Invalid login flag
    """


class NoInstanceMatched(OpenTeleException):
    """
    Invalid login flag
    """


@typing.overload
def Expects(
    condition: bool,
    message: str = None,
    done: typing.Callable[[], None] = None,
    fail: typing.Callable[[OpenTeleException], None] = None,
    silent: bool = False,
    stack_index: int = 1,
) -> bool:
    """Expect a condition to be `True`, raise an `OpenTeleException` if it's not.

    ### Arguments:
        condition (bool):
            Condition that you're expecting.

        message (str, default=None):
            Custom exception message

        done (`lambda`, default=None):
            lambda to execute when done without error

        fail (`lambda`, default=None):
            lambda to execute when the condition is False, the lambda will be execute before raising the exception.

        silent (`bool`, default=False):
            if True then it won't raise the exception, only execute fail lambda.

        stack_index (`int`, default=1):
            stack index to raise the exception with trace back to where it happens, intended for internal usage.

    ### Raises:
        `OpenTeleException`: exception
    """


@typing.overload
def Expects(
    condition: bool,
    exception: OpenTeleException = None,
    done: typing.Callable[[], None] = None,
    fail: typing.Callable[[OpenTeleException], None] = None,
    silent: bool = False,
    stack_index: int = 1,
) -> bool:

    """Expect a condition to be `True`, raise an `OpenTeleException` if it's not.

    ### Arguments:
        condition (bool):
            Condition that you're expecting.

        message (OpenTeleException, default=None):
            Custom exception.

        done (`lambda`, default=None):
            lambda to execute when done without error.

        fail (`lambda`, default=None):
            lambda to execute when the condition is False, the lambda will be execute before raising the exception.

        silent (`bool`, default=False):
            if True then it won't raise the exception, only execute fail lambda.

        stack_index (`int`, default=1):
            stack index to raise the exception with trace back to where it happens, intended for internal usage.

    ### Raises:
        `OpenTeleException`: exception
    """


def Expects(
    condition: bool,
    exception: typing.Union[OpenTeleException, str] = None,
    done: typing.Callable[[], None] = None,
    fail: typing.Callable[[OpenTeleException], None] = None,
    silent: bool = False,
    stack_index: int = 1,
) -> bool:  # nocov

    if condition:
        if done != None:
            done()
        return condition

    if isinstance(exception, str):
        exception = OpenTeleException(exception, 2)

    elif exception != None and not isinstance(exception, OpenTeleException):
        raise OpenTeleException("No instance of Expects() match the arguments given", 2)

    if exception == None:
        exception = OpenTeleException("Unexpected error", 2)

    # no raise exception
    if silent:
        if fail != None:
            fail(exception)
        return condition

    else:
        stack = inspect.stack()
        frame = stack[stack_index].frame
        tb = types.TracebackType(None, frame, frame.f_lasti, frame.f_lineno)  # type: ignore
        exception = exception.with_traceback(tb)

        if fail != None:
            fail(exception)

        raise exception


def ExpectStreamStatus(stream: QDataStream, message: str = "Could not stream data"):
    Expects(
        stream.status() == QDataStream.Status.Ok,
        stack_index=2,
        exception=QDataStreamFailed(
            "Could not read keys count from mtp authorization.", stack_index=2
        ),
    )



from __future__ import annotations
import os

import platform

from typing import Any, List, Dict, Type, TypeVar, Union, Optional
from .devices import *
from .exception import *
from .utils import *

import typing


_T = TypeVar("_T")
_RT = TypeVar("_RT")


class BaseAPIMetaClass(BaseMetaClass):
    """Super high level tactic metaclass"""

    def __new__(
        cls: Type[_T], clsName: str, bases: Tuple[type], attrs: Dict[str, Any]
    ) -> _T:

        result = super().__new__(cls, clsName, bases, attrs)
        result._clsMakePID()  # type: ignore
        result.__str__ = BaseAPIMetaClass.__str__  # type: ignore

        return result

    @sharemethod
    def __str__(glob) -> str:

        if isinstance(glob, type):
            cls = glob
            result = f"{cls.__name__} {{\n"
        else:
            cls = glob.__class__
            result = f"{cls.__name__}() = {{\n"

        for attr, val in glob.__dict__.items():

            if (
                attr.startswith(f"_{cls.__base__.__name__}__")
                or attr.startswith(f"_{cls.__name__}__")
                or attr.startswith("__")
                and attr.endswith("__")
                or type(val) == classmethod
                or callable(val)
            ):
                continue

            result += f"    {attr}: {val}\n"

        return result + "}"


class APIData(object, metaclass=BaseAPIMetaClass):
    """
    API configuration to connect to `TelegramClient` and `TDesktop`

    ### Attributes:
        api_id (`int`):
            [API_ID](https://core.telegram.org/api/obtaining_api_id#obtaining-api-id)

        api_hash (`str`):
            [API_HASH](https://core.telegram.org/api/obtaining_api_id#obtaining-api-id)

        device_model (`str`):
            Device model name

        system_version (`str`):
            Operating System version

        app_version (`str`):
            Current app version

        lang_code (`str`):
            Language code of the client

        system_lang_code (`str`):
            Language code of operating system

        lang_pack (`str`):
            Language pack

    ### Methods:
        `Generate()`: Generate random device model and system version
    """

    CustomInitConnectionList: List[Union[Type[APIData], APIData]] = []

    api_id: int = None  # type: ignore
    api_hash: str = None  # type: ignore
    device_model: str = None  # type: ignore
    system_version: str = None  # type: ignore
    app_version: str = None  # type: ignore
    lang_code: str = None  # type: ignore
    system_lang_code: str = None  # type: ignore
    lang_pack: str = None  # type: ignore

    @typing.overload
    def __init__(self, api_id: int, api_hash: str) -> None:
        pass

    @typing.overload
    def __init__(
        self,
        api_id: int,
        api_hash: str,
        device_model: str = None,
        system_version: str = None,
        app_version: str = None,
        lang_code: str = None,
        system_lang_code: str = None,
        lang_pack: str = None,
    ) -> None:
        """
        Create your own customized API

        ### Arguments:
            api_id (`int`):
                [API_ID](https://core.telegram.org/api/obtaining_api_id#obtaining-api-id)

            api_hash (`str`):
                [API_HASH](https://core.telegram.org/api/obtaining_api_id#obtaining-api-id)

            device_model (`str`, default=`None`):
                `[Device model name](API.device_model)`

            system_version (`str`, default=`None`):
                `[Operating System version](API.system_version)`

            app_version (`str`, default=`None`):
                `[Current app version](API.app_version)`

            lang_code (`str`, default=`"en"`):
                `[Language code of the client](API.app_version)`

            system_lang_code (`str`, default=`"en"`):
                `[Language code of operating system](API.system_lang_code)`

            lang_pack (`str`, default=`""`):
                `[Language pack](API.lang_pack)`

        ### Warning:
            Use at your own risk!:
                Using the wrong API can lead to your account banned.
                If the session was created using an official API, you must continue using official APIs for that session.
                Otherwise that account is at risk of getting banned.
        """

    def __init__(
        self,
        api_id: int = None,
        api_hash: str = None,
        device_model: str = None,
        system_version: str = None,
        app_version: str = None,
        lang_code: str = None,
        system_lang_code: str = None,
        lang_pack: str = None,
    ) -> None:

        Expects(
            (self.__class__ != APIData) or (api_id != None and api_hash != None),
            NoInstanceMatched("No instace of API matches the arguments"),
        )

        cls = self.get_cls()

        self.api_id = api_id if api_id else cls.api_id
        self.api_hash = api_hash if api_hash else cls.api_hash
        self.device_model = device_model if device_model else cls.device_model
        self.system_version = system_version if system_version else cls.system_version
        self.app_version = app_version if app_version else cls.app_version
        self.system_lang_code = (
            system_lang_code if system_lang_code else cls.system_lang_code
        )
        self.lang_pack = lang_pack if lang_pack else cls.lang_pack
        self.lang_code = lang_code if lang_code else cls.lang_code

        if self.device_model == None:
            system = platform.uname()

            if system.machine in ("x86_64", "AMD64"):
                self.device_model = "PC 64bit"
            elif system.machine in ("i386", "i686", "x86"):
                self.device_model = "PC 32bit"
            else:
                self.device_model = system.machine

        self._makePID()

    @sharemethod
    def copy(glob: Union[Type[_T], _T] = _T) -> _T:  # type: ignore

        cls = glob if isinstance(glob, type) else glob.__class__

        return cls(
            glob.api_id,  # type: ignore
            glob.api_hash,  # type: ignore
            glob.device_model,  # type: ignore
            glob.system_version,  # type: ignore
            glob.app_version,  # type: ignore
            glob.lang_code,  # type: ignore
            glob.system_lang_code,  # type: ignore
            glob.lang_pack,  # type: ignore
        )  # type: ignore

    @sharemethod
    def get_cls(glob: Union[Type[_T], _T]) -> Type[_T]:  # type: ignore
        return glob if isinstance(glob, type) else glob.__class__

    @sharemethod
    def destroy(glob: Union[Type[_T], _T]):  # type: ignore
        if isinstance(glob, type):
            return

        # might cause conflict, disabled for now, it won"t be a problem
        # if (API.findData(self.pid) != None):
        #     API.CustomInitConnectionList.remove(self)

    def __eq__(self, __o: APIData) -> bool:
        if not isinstance(__o, APIData):
            return False
        return self.pid == __o.pid

    def __del__(self):
        self.destroy()

    @classmethod
    def _makePIDEnsure(cls) -> int:
        while True:
            pid = int.from_bytes(os.urandom(8), "little")
            if cls.findData(pid) == None:
                break
        return pid

    @classmethod
    def _clsMakePID(cls: Type[APIData]):
        cls.pid = cls._makePIDEnsure()
        cls.CustomInitConnectionList.append(cls)

    def _makePID(self):
        self.pid = self.get_cls()._makePIDEnsure()
        self.get_cls().CustomInitConnectionList.append(self)

    @classmethod
    def Generate(cls: Type[_T], unique_id: str = None) -> _T:
        """
        Generate random device model and system version

        ### Arguments:
            unique_id (`str`, default=`None`):
                The unique ID to generate - can be anything.\\
                This will be used to ensure that it will generate the same data everytime.\\
                If not set then the data will be randomized each time we runs it.
        
        ### Raises:
            `NotImplementedError`: Not supported for web browser yet

        ### Returns:
            `APIData`: Return a copy of the api with random device data

        ### Examples:
            Create a `TelegramClient` with custom API:
        ```python
            api = API.TelegramIOS.Generate(unique_id="new.session")
            client = TelegramClient(session="new.session" api=api)
            client.start()
        ```
        """
        if cls == API.TelegramAndroid or cls == API.TelegramAndroidX:
            deviceInfo = AndroidDevice.RandomDevice(unique_id)

        elif cls == API.TelegramIOS:
            deviceInfo = iOSDeivce.RandomDevice(unique_id)

        elif cls == API.TelegramMacOS:
            deviceInfo = macOSDevice.RandomDevice(unique_id)

        # elif cls == API.TelegramWeb_K or cls == API.TelegramWeb_Z or cls == API.Webogram:
        else:
            raise NotImplementedError(
                f"{cls.__name__} device not supported for randomize yet"
            )

        return cls(device_model=deviceInfo.model, system_version=deviceInfo.version)

    @classmethod
    def findData(cls: Type[_T], pid: int) -> Optional[_T]:
        for x in cls.CustomInitConnectionList:  # type: ignore
            if x.pid == pid:
                return x
        return None


class API(BaseObject):
    """
    #### Built-in templates for Telegram API
    - **`opentele`** offers the ability to use **`official APIs`**, which are used by `official apps`.
    - According to [Telegram TOS](https://core.telegram.org/api/obtaining_api_id#using-the-api-id): *all accounts that sign up or log in using unofficial Telegram API clients are automatically put under observation to avoid violations of the Terms of Service*.
    - It also uses the **[lang_pack](https://core.telegram.org/method/initConnection)** parameter, of which [telethon can't use](https://github.com/LonamiWebs/Telethon/blob/master/telethon/_client/telegrambaseclient.py#L192) because it's for official apps only.
    - Therefore, **there are no differences** between using `opentele` and `official apps`, the server can't tell you apart.
    - You can use `TelegramClient.PrintSessions()` to check this out.

    ### Attributes:
        TelegramDesktop (`API`):
            Official Telegram for Desktop (Windows, macOS and Linux) [View on GitHub](https://github.com/telegramdesktop/tdesktop)

        TelegramAndroid (`API`):
            Official Telegram for Android [View on GitHub](https://github.com/DrKLO/Telegram)

        TelegramAndroidX (`API`):
            Official TelegramX for Android [View on GitHub](https://github.com/DrKLO/Telegram)

        TelegramIOS (`API`):
            Official Telegram for iOS [View on GitHub](https://github.com/TelegramMessenger/Telegram-iOS)

        TelegramMacOS (`API`):
            Official Telegram-Swift For MacOS [View on GitHub](https://github.com/overtake/TelegramSwift)

        TelegramWeb_Z (`API`):
            Default Official Telegram Web Z For Browsers [View on GitHub](https://github.com/Ajaxy/telegram-tt) | [Visit on Telegram](https://web.telegram.org/z/)

        TelegramWeb_K (`API`):
            Official Telegram Web K For Browsers [View on GitHub](https://github.com/morethanwords/tweb) | [Visit on Telegram](https://web.telegram.org/k/)

        Webogram (`API`):
            Old Telegram For Browsers [View on GitHub](https://github.com/zhukov/webogram) | [Vist on Telegram](https://web.telegram.org/?legacy=1#/im)
    """

    class TelegramDesktop(APIData):
        """
        Official Telegram for Desktop (Windows, macOS and Linux)
        [View on GitHub](https://github.com/telegramdesktop/tdesktop)

        ### Attributes:
            api_id (`int`)           : `2040`
            api_hash (`str`)         : `"b18441a1ff607e10a989891a5462e627"`
            device_model (`str`)     : `"Desktop"`
            system_version (`str`)   : `"Windows 10"`
            app_version (`str`)      : `"3.4.3 x64"`
            lang_code (`str`)        : `"en"`
            system_lang_code (`str`) : `"en-US"`
            lang_pack (`str`)        : `"tdesktop"`

        ### Methods:
            `Generate()`: Generate random device data for `Windows`, `macOS` and `Linux`
        """

        api_id = 2040
        api_hash = "b18441a1ff607e10a989891a5462e627"
        device_model = "Desktop"
        system_version = "Windows 10"
        app_version = "3.4.3 x64"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "tdesktop"

        @typing.overload
        @classmethod
        def Generate(
            cls: Type[_T], system: str = "windows", unique_id: str = None
        ) -> _T:
            """
            Generate random TelegramDesktop devices
            ### Arguments:
                system (`str`, default=`"random"`):
                    Which OS to generate, either `"windows"`, `"macos"`, or `"linux"`.\\
                    Default is `None` or `"random"` -  which means it will be selected randomly.
            
                unique_id (`str`, default=`None`):
                    The unique ID to generate - can be anything.\\
                    This ID will be used to ensure that it will generate the same data every single time.\\
                    If not set then the data will be randomized each time we runs it.
            
            ### Returns:
                `APIData`: Return a copy of the api with random device data
            
            ### Examples:
                Save a telethon session to tdata:
            ```python
                # unique_id will ensure that this data will always be the same (per unique_id).
                # You can use the session file name, or user_id as a unique_id.
                # If unique_id isn't specify, the device data will be randomized each time we runs it.
                oldAPI = API.TelegramDesktop.Generate(system="windows", unique_id="old.session")
                oldclient = TelegramClient("old.session", api=oldAPI)
                await oldClient.connect()

                # We can safely CreateNewSession with a different API.
                # Be aware that you should not use UseCurrentSession with a different API than the one that first authorized it.
                # You can print(newAPI) to see what it had generated.
                newAPI = API.TelegramDesktop.Generate("macos", "new_tdata")
                tdesk = oldclient.ToTDesktop(oldclient, flag=CreateNewSession, api=newAPI)

                # Save the new session to a folder named "new_tdata"
                tdesk.SaveTData("new_tdata")
            ```
            """

        @typing.overload
        @classmethod
        def Generate(cls: Type[_T], system: str = "macos", unique_id: str = None) -> _T:
            pass

        @typing.overload
        @classmethod
        def Generate(cls: Type[_T], system: str = "linux", unique_id: str = None) -> _T:
            pass

        @typing.overload
        @classmethod
        def Generate(
            cls: Type[_T], system: str = "random", unique_id: str = None
        ) -> _T:
            pass

        @classmethod
        def Generate(cls: Type[_T], system: str = None, unique_id: str = None) -> _T:

            validList = ["windows", "macos", "linux"]
            if system == None or system not in validList:
                system = SystemInfo._hashtovalue(
                    SystemInfo._strtohashid(unique_id), validList
                )

            system = system.lower()

            if system == "windows":
                deviceInfo = WindowsDevice.RandomDevice(unique_id)

            elif system == "macos":
                deviceInfo = macOSDevice.RandomDevice(unique_id)

            else:
                deviceInfo = LinuxDevice.RandomDevice(unique_id)

            return cls(device_model=deviceInfo.model, system_version=deviceInfo.version)

    class TelegramAndroid(APIData):
        """
        Official Telegram for Android
        [View on GitHub](https://github.com/DrKLO/Telegram)

        ### Attributes:
            api_id (`int`)           : `6`
            api_hash (`str`)         : `"eb06d4abfb49dc3eeb1aeb98ae0f581e"`
            device_model (`str`)     : `"Samsung SM-G998B"`
            system_version (`str`)   : `"SDK 31"`
            app_version (`str`)      : `"8.4.1 (2522)"`
            lang_code (`str`)        : `"en"`
            system_lang_code (`str`) : `"en-US"`
            lang_pack (`str`)        : `"android"`
        """

        api_id = 6
        api_hash = "eb06d4abfb49dc3eeb1aeb98ae0f581e"
        device_model = "Samsung SM-G998B"
        system_version = "SDK 31"
        app_version = "8.4.1 (2522)"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "android"

    class TelegramAndroidX(APIData):
        """
        Official TelegramX for Android
        [View on GitHub](https://github.com/DrKLO/Telegram)

        ### Attributes:
            api_id (`int`)           : `21724`
            api_hash (`str`)         : `"3e0cb5efcd52300aec5994fdfc5bdc16"`
            device_model (`str`)     : `"Samsung SM-G998B"`
            system_version (`str`)   : `"SDK 31"`
            app_version (`str`)      : `"8.4.1 (2522)"`
            lang_code (`str`)        : `"en"`
            system_lang_code (`str`) : `"en-US"`
            lang_pack (`str`)        : `"android"`
        """

        api_id = 21724
        api_hash = "3e0cb5efcd52300aec5994fdfc5bdc16"
        device_model = "Samsung SM-G998B"
        system_version = "SDK 31"
        app_version = "8.4.1 (2522)"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "android"

    class TelegramIOS(APIData):
        """
        Official Telegram for iOS
        [View on GitHub](https://github.com/TelegramMessenger/Telegram-iOS)

        ### Attributes:
            api_id (`int`)           : `10840`
            api_hash (`str`)         : `"33c45224029d59cb3ad0c16134215aeb"`
            device_model (`str`)     : `"iPhone 13 Pro Max"`
            system_version (`str`)   : `"14.8.1"`
            app_version (`str`)      : `"8.4"`
            lang_code (`str`)        : `"en"`
            system_lang_code (`str`) : `"en-US"`
            lang_pack (`str`)        : `"ios"`
        """

        # api_id           = 8
        # api_hash         = "7245de8e747a0d6fbe11f7cc14fcc0bb"
        api_id = 10840
        api_hash = "33c45224029d59cb3ad0c16134215aeb"
        device_model = "iPhone 13 Pro Max"
        system_version = "14.8.1"
        app_version = "8.4"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "ios"

    class TelegramMacOS(APIData):
        """
        Official Telegram-Swift For MacOS
        [View on GitHub](https://github.com/overtake/TelegramSwift)

        ### Attributes:
            api_id (`int`)           : `2834`
            api_hash (`str`)         : `"68875f756c9b437a8b916ca3de215815"`
            device_model (`str`)     : `"MacBook Pro"`
            system_version (`str`)   : `"macOS 12.0.1"`
            app_version (`str`)      : `"8.4"`
            lang_code (`str`)        : `"en"`
            system_lang_code (`str`) : `"en-US"`
            lang_pack (`str`)        : `"macos"`
        """

        api_id = 2834
        api_hash = "68875f756c9b437a8b916ca3de215815"
        # api_id = 9                                    |
        # api_hash = "3975f648bb682ee889f35483bc618d1c" | Telegram for macOS uses this api, but it"s unofficial api, why?
        device_model = "MacBook Pro"
        system_version = "macOS 12.0.1"
        app_version = "8.4"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "macos"

    class TelegramWeb_Z(APIData):
        """
        Default Official Telegram Web Z For Browsers
        [View on GitHub](https://github.com/Ajaxy/telegram-tt) | [Visit on Telegram](https://web.telegram.org/z/)

        ### Attributes:
            api_id (`int`)           : `2496`
            api_hash (`str`)         : `"8da85b0d5bfe62527e5b244c209159c3"`
            device_model (`str`)     : `"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36"`
            system_version (`str`)   : `"Windows"`
            app_version (`str`)      : `"1.28.3 Z"`
            lang_code (`str`)        : `"en"`
            system_lang_code (`str`) : `"en-US"`
            lang_pack (`str`)        : `""`
        """

        api_id = 2496
        api_hash = "8da85b0d5bfe62527e5b244c209159c3"
        device_model = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36"
        system_version = "Windows"
        app_version = "1.28.3 Z"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = ""  # I don"t understand why Telegram Z doesn"t use langPack
        # You can read its source here: https://github.com/Ajaxy/telegram-tt/blob/f7bc473d51c0fe3a3e8b22678b62d2360225aa7c/src/lib/gramjs/client/TelegramClient.js#L131

    class TelegramWeb_K(APIData):
        """
        Official Telegram Web K For Browsers
        [View on GitHub](https://github.com/morethanwords/tweb) | [Visit on Telegram](https://web.telegram.org/k/)

        ### Attributes:
            api_id (`int`)           : `2496`
            api_hash (`str`)         : `"8da85b0d5bfe62527e5b244c209159c3"`
            device_model (`str`)     : `"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36"`
            system_version (`str`)   : `"Win32"`
            app_version (`str`)      : `"1.0.1 K"`
            lang_code (`str`)        : `"en"`
            system_lang_code (`str`) : `"en-US"`
            lang_pack (`str`)        : `"macos"`
        """

        api_id = 2496
        api_hash = "8da85b0d5bfe62527e5b244c209159c3"
        device_model = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36"
        system_version = "Win32"
        app_version = "1.0.1 K"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "macos"  # I"m totally confused, why macos? https://github.dev/morethanwords/tweb/blob/26582590e647766f5890c79e1611c54c1e6e800c/src/config/app.ts#L23

    class Webogram(APIData):
        """
        Old Telegram For Browsers
        [View on GitHub](https://github.com/zhukov/webogram) | [Vist on Telegram](https://web.telegram.org/?legacy=1#/im)

        ### Attributes:
            api_id (`int`)           : `2496`
            api_hash (`str`)         : `"8da85b0d5bfe62527e5b244c209159c3"`
            device_model (`str`)     : `"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36"`
            system_version (`str`)   : `"Win32"`
            app_version (`str`)      : `"0.7.0"`
            lang_code (`str`)        : `"en"`
            system_lang_code (`str`) : `"en-US"`
            lang_pack (`str`)        : `""`
        """

        api_id = 2496
        api_hash = "8da85b0d5bfe62527e5b244c209159c3"
        device_model = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36"
        system_version = "Win32"
        app_version = "0.7.0"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = ""  # The same problem as TelegramWeb_K, as TelegramWeb_K was built on Webogram


class LoginFlag(int):
    """
    Login flag for converting sessions between `TDesktop` and `TelegramClient`.

    ### Attributes:
        UseCurrentSession (LoginFlag): Use the current session.
        CreateNewSession (LoginFlag): Create a new session.

    ### Related:
        - `TDesktop.ToTelethon()`
        - `TDesktop.FromTelethon()`
        - `TelegramClient.ToTDesktop()`
        - `TelegramClient.FromTDesktop()`

    """


class UseCurrentSession(LoginFlag):
    """
    Use the current session.
    - Convert an already-logged in session of `Telegram Desktop` to `Telethon` and vice versa.
    - The "session" is just an 256-bytes `AuthKey` that get stored in `tdata folder` or Telethon `session files` [(under sqlite3 format)](https://docs.telethon.dev/en/latest/concepts/sessions.html?highlight=sqlite3#what-are-sessions).
    - `UseCurrentSession`'s only job is to read this key and convert it to one another.

    ### Warning:
        Use at your own risk!:
            You should only use the same consistant API through out the session.
            Don't use a same session with multiple different APIs, you might be banned.


    """


class CreateNewSession(LoginFlag):
    """
    Create a new session.
    - Use the `current session` to authorize the `new session` by [Login via QR code](https://core.telegram.org/api/qr-login).
    - This works just like when you signing into `Telegram` using `QR Login` on mobile devices.
    - Although `Telegram Desktop` doesn't let you authorize other sessions via `QR Code` *(or it doesn't have that feature)*, it is still available across all platforms `(``[APIs](API)``)`.

    ### Done:
        Safe to use:
            You can always use `CreateNewSessions` with any APIs, it can be different from the API that originally created the session.
    """



import inspect
import os

import time
import atexit

import typing as t

IS_DEBUG_MODE = False
# IS_DEBUG_MODE = True

if IS_DEBUG_MODE:
    from rich import print  # pragma: no cover

    _F = t.TypeVar("_F")
    _T = t.TypeVar("_T")
    _R = t.TypeVar("_R")

    class DebugInfo(object):
        def __init__(self) -> None:
            self.list: t.List[t.List[str, int, int, int]] = []

        def add(self, name, total_called, total_time) -> None:
            self.list.append(
                (
                    name,
                    total_called,
                    total_time,
                    round(total_time / total_called * 1000, 4),
                )
            )

        def on_exit(self):
            self.list = sorted(self.list, key=lambda item: item[3])

            for item in self.list:

                name = item[0]
                total_called = item[1]
                total_time = item[2]
                avg_time = item[3]

                text = name.ljust(40)
                text += f"called {total_called}".ljust(15)
                text += f"average time: {avg_time} ms".ljust(25)
                print(text)

    dbgInfo = DebugInfo()
    atexit.register(dbgInfo.on_exit)

    class DebugMethod(type):  # pragma: no cover
        def __get__(self, obj, cls):
            self.__owner__ = obj if obj else cls
            return self

        def __call2__(self, *args, **kwargs):
            begin = time.perf_counter()
            result = self.__fget__(*args, **kwargs)  # type: ignore
            diff = time.perf_counter() - begin

            return result, diff

        def __call__(self, *args, **kwargs) -> t.Any:

            # stack = inspect.stack()[1]

            # location = f"{os.path.basename(stack.filename)[:-3]}"
            # lineno = f"{stack.lineno}."

            # if len(location) >= 15:
            #     location = ".." + location[-15:]

            # location = location.ljust(15) + "| " + lineno.ljust(6)

            # context = stack.code_context[0][:-1].strip()  # type: ignore
            # print(location + context, f"took {diff}ms")

            if hasattr(self, "__owner__"):
                result, diff = DebugMethod.__call2__(
                    self, self.__owner__, *args, **kwargs
                )
            else:
                result, diff = DebugMethod.__call2__(self, *args, **kwargs)

            self.__total_called += 1
            self.__total_time += diff

            return result

        def __getfullname(self):
            if hasattr(self, "__ownername__"):
                if self.__fname__ == "__init__":
                    return f"{self.__ownername__}()"

                return f"{self.__ownername__}.{self.__fname__}()"
            else:
                return f"{self.__fget__.__name__}()"

        def on_exit(self):
            if self.__total_called > 0:
                dbgInfo.add(
                    DebugMethod.__getfullname(self),
                    self.__total_called,
                    self.__total_time,
                )

        def __set_name__(self, owner, name):
            self.__owner__: str = owner
            self.__ownername__: str = owner.__name__
            self.__fname__: str = name
            if self.__fname__.startswith(f"_{self.__ownername__}__"):
                self.__fname__ = self.__fname__[len(self.__ownername__) + 1 :]

        def __new__(cls, decorated_func: _F) -> _F:

            firstdct = dict(decorated_func.__dict__)
            for i, x in cls.__dict__.items():
                firstdct[i] = x

            result = type.__new__(
                cls,
                decorated_func.__class__.__name__,
                decorated_func.__class__.__bases__,
                firstdct,
            )

            result.__fget__ = decorated_func
            result.__total_time = 0
            result.__total_called = 0
            atexit.register(result.on_exit, result)
            return result

    def parse_arg(value) -> str:  # pragma: no cover
        if isinstance(value, type):
            return value.__name__
        elif isinstance(value, str):
            return f"'{value}'"
        elif isinstance(value, int):
            return f"{value}"
        elif isinstance(value, object):
            return value.__class__.__name__ + "(...)"
        return value

    class runtime(type):  # pragma: no cover
        def __get__(self, obj, cls):
            self.__owner__ = obj if obj else cls
            return self

        def __call__(self, *args, **kwargs) -> t.Any:

            begin = time.perf_counter()
            result = self.__fget__(self.__owner__, *args, **kwargs)  # type: ignore
            diff = round((time.perf_counter() - begin) * 1000, 2)
            print(f"{self.__fname__} took {diff}ms")
            return result

        def __set_name__(self, owner, name):
            print("set_name")
            self.__owner__ = owner
            self.__ownername__ = owner.__name__
            self.__fname__ = name

        def __new__(cls, decorated_func):

            firstdct = dict(decorated_func.__dict__)
            for i, x in cls.__dict__.items():
                firstdct[i] = x

            result = type.__new__(
                cls,
                decorated_func.__name__,
                decorated_func.__class__.__bases__,
                firstdct,
            )
            result.__fget__ = decorated_func  # type: ignore
            return result



from __future__ import annotations


from .configs import *
from . import shared as td

import logging
from telethon.network.connection.connection import Connection
from telethon.network.connection.tcpfull import ConnectionTcpFull
from telethon.sessions.abstract import Session

# if TYPE_CHECKING:
#     from . import *


class TDesktop(BaseObject):
    """
    Telegram Desktop client.
    
    A client can have multiple accounts, up to 3 - according to official Telegram Desktop client.

    ### Attributes:
        api (`API`):
            The API this client is using.
        
        accountsCount (`int`):
            The numbers of accounts in this client.

        accounts (`List[Account]`):
            List of accounts in this client.

        mainAccount (`Account`):
            The main account of this client.
        
        basePath (`str`):
            The path to `tdata folder`.

        passcode (`str`):
            Passcode of the client, the same as Local Passcode on `Telegram Desktop`.\\
            Use to encrypt and decrypt `tdata files`.

        AppVersion (`int`):
            App version of the client.
        
        kMaxAccounts (`int`):
            See `kMaxAccounts`.

        keyFile (`str`):
            See `keyFile`.
        
        kDefaultKeyFile (`str`):
            Default value for `keyFile`.

        kPerformanceMode (`bool`):
            Performance mode. When enabled, `SavaTData()` will be 200x faster. 
            See `kPerformanceMode`.
    
    ### Methods:
        LoadTData():
            Load the client from `tdata folder`. \\
            Use this if you didn't set the `basePath` when initializing the client.

        SaveTData():
            Save the client session to `tdata folder` - which can be used by `Telegram Desktop`.

        isLoaded():
            Return `True` if the client has successfully loaded accounts from `tdata` or `TelegramClient`.

        ToTelethon():
            Convert this session to `TelegramClient`.

        FromTelethon():
            Create a new session from `TelegramClient`.

        PerformanceMode():
            Enable/disable performance mode. When enabled, `SavaTData()` will be 5000x faster.

    """

    @typing.overload
    def __init__(self) -> None:
        pass

    @typing.overload
    def __init__(
        self,
        basePath: str = None,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
    ) -> None:
        pass

    @typing.overload
    def __init__(
        self,
        basePath: str = None,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        passcode: str = None,
        keyFile: str = None,
    ) -> None:
        pass

    def __init__(
        self,
        basePath: str = None,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        passcode: str = None,
        keyFile: str = None,
    ) -> None:
        """
        Initialize a `TDesktop` client

        ### Arguments:
            basePath (`str`, default=`None`):
                The path to the `tdata folder`.
                If the path doesn't exists or its data is corrupted, a new instance will be created.

            api (`API`, default=`TelegramDesktop`):
                Which API to use. Read more `[here](API)`.

            passcode (str, default=`None`):
                The passcode for tdata, same as the Local Passcode on `Telegram Desktop`.

            keyFile (str, default="data"):
                See `keyFile`.
        """
        self.__accounts: typing.List[td.Account] = []
        self.__basePath = basePath
        self.__keyFile = keyFile if (keyFile != None) else TDesktop.kDefaultKeyFile
        self.__passcode = passcode if (passcode != None) else str("")
        self.__passcodeBytes = self.__passcode.encode("ascii")
        self.__mainAccount: Optional[td.Account] = None
        self.__active_index = -1
        self.__passcodeKey = None
        self.__localKey = None
        self.__AppVersion = None
        self.__isLoaded = False
        self.__api = api.copy()

        if basePath != None:
            self.__basePath = td.Storage.GetAbsolutePath(basePath)
            self.LoadTData()

    def isLoaded(self) -> bool:
        """
        Return `True` if the client has successfully loaded accounts from `tdata` or `TelegramClient`
        """
        return self.__isLoaded

    def LoadTData(
        self, basePath: str = None, passcode: str = None, keyFile: str = None
    ):
        """
        Loads accounts from `tdata folder`

        ### Arguments:
            basePath (`str`, default=`None`):
                The path to the folder.

            passcode (`str`, default=`None`):
                Read more `[here](passcode)`.

            keyFile (`str`, default=`None`):
                Read more `[here](keyFile)`.

        ### Raises:
            `TDataBadDecryptKey`: The `tdata folder` is password-encrypted, please the set the argument `passcode` to decrypt it.

        ### Warning:
            This function is not recommended to use:
                You should load tdata using `TDesktop(basePath="path")`.
                Don't manually load tdata using this function, bugs might pop up out of nowhere.

        ### Examples:
        ```python
            # Using the API that we've generated before. Please refer to method API.Generate() to learn more.
            oldAPI = API.TelegramDesktop.Generate(system="windows", unique_id="old.session")
            oldclient = TelegramClient("old.session", api=oldAPI)
            await oldClient.connect()

            # We can safely use CreateNewSession with a different API.
            # Be aware that you should not use UseCurrentSession with a different API than the one that first authorized it.
            newAPI = API.TelegramAndroid.Generate("new_tdata")
            tdesk = await TDesktop.FromTelethon(oldclient, flag=CreateNewSession, api=newAPI)

            # Save the new session to a folder named "new_tdata"
            tdesk.SaveTData("new_tdata")
        ```
        """

        if basePath == None:
            basePath = self.basePath

        Expects(basePath != None and basePath != "", "No folder provided to load tdata")

        if keyFile != None and self.__keyFile != keyFile:
            self.__keyFile = keyFile

        if passcode != None and self.__passcode != passcode:
            self.__passcode = passcode
            self.__passcodeBytes = passcode.encode("ascii")

        try:
            self.__loadFromTData()

        except OpenTeleException as e:
            if isinstance(e, TDataBadDecryptKey):
                if self.passcode == "":
                    raise TDataBadDecryptKey(
                        "The tdata folder is password-encrypted, please the set the argument 'passcode' to decrypt it"
                    )
                else:
                    raise TDataBadDecryptKey(
                        "Failed to decrypt tdata folder because of invalid passcode"
                    )
            else:
                raise e

        Expects(self.isLoaded(), "Failed to load? Something went seriously wrong")

    def SaveTData(
        self, basePath: str = None, passcode: str = None, keyFile: str = None
    ) -> bool:
        """
        Save the client session to a folder.

        ### Arguments:
            basePath (`str`, default=None):
                Path to the folder\\
                If None then the data will be saved at the basePath given at creation.

            passcode (`str`, default=`None`):
                Read more `[here](passcode)`.
        
            keyFile (`str`, default=`None`):
                Read more `[here](keyFile)`.

        ### Examples:
            Save a telethon session to tdata:
        ```python
            # Using the API that we've generated before. Please refer to method API.Generate() to learn more.
            oldAPI = API.TelegramDesktop.Generate(system="windows", unique_id="old.session")
            oldclient = TelegramClient("old.session", api=oldAPI)
            await oldClient.connect()

            # We can safely CreateNewSession with a different API.
            # Be aware that you should not use UseCurrentSession with a different API than the one that first authorized it.
            newAPI = API.TelegramAndroid.Generate("new_tdata")
            tdesk = await TDesktop.FromTelethon(oldclient, flag=CreateNewSession, api=newAPI)

            # Save the new session to a folder named "new_tdata"
            tdesk.SaveTData("new_tdata")
        ```
        """
        if basePath == None:
            basePath = self.basePath

        self.__keyFile = (
            keyFile if (keyFile != None and self.keyFile != keyFile) else self.keyFile
        )

        Expects(basePath != None and basePath != "", "No folder provided to save tdata")

        if passcode != None and self.__passcode != passcode:
            self.__passcode = passcode
            self.__passcodeBytes = passcode.encode("ascii")
            self.__isLoaded = False  # to generate new localKey

        if not self.isLoaded():
            self.__generateLocalKey()
        Expects(self.isLoaded(), "Failed to load? Something went seriously wrong")

        try:

            basePath = td.Storage.GetAbsolutePath(basePath)
            if not self.basePath:
                self.__basePath = basePath

            self.__writeAccounts(basePath, keyFile)
            return True

        except OpenTeleException as e:
            raise TDataSaveFailed("Could not save tdata, something went wrong") from e

    def __writeAccounts(self, basePath: str, keyFile: str = None) -> None:
        # Intended for internal usage only

        Expects(len(self.accounts) > 0)
        Expects(basePath != None and basePath != "", "No folder provided to save tdata")

        for account in self.accounts:
            account._writeData(basePath, keyFile)

        key = td.Storage.FileWriteDescriptor("key_" + self.keyFile, basePath)
        key.writeData(self.__passcodeKeySalt)
        key.writeData(self.__passcodeKeyEncrypted)

        keySize = sizeof(int32) + sizeof(int32) * len(self.accounts)
        keyData = td.Storage.EncryptedDescriptor(keySize)
        keyData.stream.writeInt32(len(self.accounts))

        for account in self.accounts:
            keyData.stream.writeInt32(account.index)

        keyData.stream.writeInt32(self.__active_index)
        key.writeEncrypted(keyData, self.__localKey)  # type: ignore
        key.finish()

    def __generateLocalKey(self) -> None:
        # Intended for internal usage only

        Expects(not self.isLoaded())

        if self.kPerformanceMode and len(self.__passcodeBytes) == 0:

            self.__localKey = td.AuthKey(
                b"\xd8\x74\x59\x44\x51\x9e\x0d\x2d\x71\x30\x9d\x6c\x8d\x27\x2d\xc6\x49\x48\xf5\xe3\xeb\xa7\x68\x53\x24\xd5\xc6\x91\xad\x81\x0c\x20"
                b"\x3b\x31\xd1\x9d\x29\xae\xd6\xac\x33\xc0\x14\xbe\x6e\x09\x84\x32\x93\xf6\xfa\x32\xdb\xe4\x2b\x6a\x04\xe0\x04\x81\xfa\xe9\x95\x11"
                b"\x4c\xaf\x63\x42\xbd\x98\xe9\x6d\x29\x3d\xd0\x62\xc4\x58\x68\x9b\x3a\xbd\x23\xa5\xcf\x23\x0c\x75\x52\x7c\x05\xbf\x5f\x90\xf3\x8c"
                b"\xd9\x39\x52\xcf\x61\xaa\xac\x1c\xfe\xaa\xe4\x60\x85\x92\xe3\x63\xde\xd3\x5f\x8d\x8c\x45\x23\x4d\xef\x53\x23\x1d\xec\xb3\x55\x92"
                b"\xaf\xc4\x0d\x06\x01\xbb\xed\x11\x09\x09\x69\xf7\x4d\x9a\xb0\xcc\x97\x82\x75\x46\xf4\x41\x24\x2d\x2c\xfb\x8e\x05\xa0\x61\x0e\x97"
                b"\x66\x9c\x0d\xa1\xad\xcc\xb5\x6e\x39\xe1\x0c\x69\xe2\x94\x23\x87\xff\x49\x22\xf8\xc5\x5d\xcb\x88\x90\xe3\x45\xef\x31\x82\x66\xf4"
                b"\xb3\x83\x14\x30\xea\x21\x0c\x86\x3c\x17\x62\x4c\x04\x94\xcd\xea\xd8\x1f\x52\x34\x30\xb5\xf7\x4c\x15\xda\x32\x3d\x76\x6b\xd0\x1c"
                b"\xb5\xb8\x8b\x9d\x2a\x73\x1f\x6d\x85\x33\x80\xad\x30\x6a\x86\x47\xfa\x61\x4c\xc4\x01\x7f\x08\x90\x2c\x1e\x1f\x99\x7e\xe1\x2e\x3c"
            )
            self.__passcodeKeySalt = QByteArray(
                b"\xae\xd1\xe0\x82\x99\x42\x81\xd9\x75\x76\x0e\x72\x95\x60\xd2\xc8\xd0\x08\xf2\xa9\xdd\x3f\xf4\xd8\x32\x45\xe2\x2e\xed\xb6\x67\x16"
            )
            self.__passcodeKey = td.AuthKey(
                b"\x27\x3e\x64\x83\xb3\x13\xc9\xdb\xc4\xac\xd9\x17\x38\x64\xc0\x42\x2f\x17\x28\x81\xbf\xe1\xc6\x64\x9c\xa5\x53\x86\x54\xd0\xbd\x6e"
                b"\xbc\xfb\xcc\x8d\xab\xbb\x66\x97\x17\xce\x53\xb4\x1c\xa8\xaf\xbf\x9d\x15\xc2\x3f\xa2\xb0\x6a\x6e\x16\x6c\xc6\x1f\x62\xe9\x98\x05"
                b"\x58\xcd\x42\xcf\x10\x22\xe0\x5b\x46\x68\x11\xea\x29\xe9\x35\x9f\xc1\xf9\x46\x68\xe5\xc8\x51\x55\xbc\xe6\x38\xfe\x7a\xa4\x2a\xa8"
                b"\x80\xca\x83\xe0\xd2\xe5\x34\x19\x8a\x0a\x89\x01\x39\xc8\xf6\x79\xa5\x7d\x3b\xa0\x6d\x75\xfe\x5f\xaa\x6e\x25\x01\x01\xa5\x8b\xc6"
                b"\xe4\x2b\x96\x96\x60\x4d\xe1\xe3\x4a\xe4\x0f\x6b\xba\x14\x4a\x28\x4f\x3a\xd8\x84\x32\x53\xec\x9b\x39\x71\x86\x3a\x2c\x40\x92\x08"
                b"\xc2\x56\x39\x67\xb3\x58\x7e\x50\x9b\x42\xa4\x2a\x60\x40\xd2\x3f\xf6\x96\xad\x55\x2a\x24\x00\x84\xfa\x3f\x95\x02\x40\xf4\x99\xb2"
                b"\x3c\xd6\xd2\x7e\x70\x10\xcb\xde\x07\xda\xae\x06\x67\xa7\xdf\x8a\x51\x15\xa6\x0b\x26\x5c\x58\xf5\xd9\x29\x1f\x7a\x98\xfc\x3c\x60"
                b"\x1e\x2a\x4a\x32\xf1\x88\x1b\x82\x18\xc8\x55\x23\x9d\x7b\x53\x29\x59\x60\x9e\x6a\xb5\x2e\x48\xad\x69\x1c\x25\x83\xb5\x66\xc8\xf9"
            )
            self.__passcodeKeyEncrypted = QByteArray(
                b"\x97\xdf\x0c\xd2\xe3\x10\x91\x49\xb7\x7b\x52\x87\x99\x4d\x9c\x1c\xa2\x40\xc5\x1e\x87\x48\x8e\x79"
                b"\xdd\x02\x9b\xea\x65\xfb\x9d\x27\x89\xbb\x5a\xbc\xfe\x65\xe8\x71\xd7\x52\xbd\x93\x8d\x83\x31\x3c"
                b"\x79\x4c\x89\x93\xa7\x34\xce\x12\x16\xf2\xe6\x60\x47\x3f\x31\x43\xaf\x9a\x33\x36\x10\xa1\x79\x95"
                b"\x87\x6e\x17\x21\xce\x1f\x61\x1d\x1c\x69\xd8\xc1\xa2\xf5\x9f\x94\x93\x11\x97\x04\x27\x4e\x2c\xb5"
                b"\xf3\x6c\x20\xdf\x43\x9d\x15\x6d\xef\xf7\xa3\x43\x71\xdc\x44\xbc\x86\xf8\x73\x0c\xeb\xf9\xb0\x28"
                b"\xeb\x7a\x1e\xd6\x62\x1d\x99\xad\xb6\x2b\x3b\x2c\xf2\x29\x5d\xbb\xb2\x4b\xf1\x32\xd3\x7f\xff\xc1"
                b"\x7a\x0b\xdc\xcc\x84\xbb\xea\x6e\xa3\x47\x37\xa2\x36\xb5\x82\x48\xa7\xab\x4c\x14\x36\x3c\x20\x54"
                b"\x1c\xb4\x53\x38\x67\x7f\x33\x97\x82\xb2\x05\xe3\x55\x18\x96\x58\xdd\x45\xea\x3e\x80\x05\xf8\x51"
                b"\x14\x8e\x7e\x15\xf4\x31\x90\x4f\xa7\x9c\x68\x27\xee\x42\x6d\x3a\xb9\xcb\xa9\x36\xeb\x33\xd4\x85"
                b"\xdb\x88\xa6\xf0\xff\x97\x22\xa6\xd6\x2f\xf7\x88\x34\x7e\x27\xc8\x2e\x9e\x13\x9e\xb0\x3a\xe5\x21"
                b"\x53\x9b\xf3\xd3\x63\xb4\xba\xea\x76\xe5\xe8\x84\xcf\x66\xfe\x6b\xcd\x8a\x9e\x08\x9d\x36\x40\x5d"
                b"\xb9\x9d\x01\xdb\x20\x46\x4f\xb6\xca\xbb\xdc\xe4\xf6\x7e\x4e\xc3\x74\x2f\x91\x3a\x1d\xd2\xda\xc5"
            )

        else:
            LocalEncryptSaltSize = 32

            _pass = td.Storage.RandomGenerate(td.AuthKey.kSize)
            _salt = td.Storage.RandomGenerate(LocalEncryptSaltSize)
            self.__localKey = td.Storage.CreateLocalKey(_salt, _pass)

            self.__passcodeKeySalt = td.Storage.RandomGenerate(
                LocalEncryptSaltSize
            )  # LocalEncryptSaltSize = 32
            self.__passcodeKey = td.Storage.CreateLocalKey(
                self.__passcodeKeySalt, QByteArray(self.__passcodeBytes)
            )

            passKeyData = td.Storage.EncryptedDescriptor(td.AuthKey.kSize)
            self.__localKey.write(passKeyData.stream)

            self.__passcodeKeyEncrypted = td.Storage.PrepareEncrypted(
                passKeyData, self.__passcodeKey
            )

            # set new local key for self.accounts
            for account in self.accounts:
                account.localKey = self.localKey

        self.__isLoaded = True

    def _addSingleAccount(self, account: td.Account):
        # Intended for internal usage only

        # Expects(self.isLoaded(), "Could not add account because i haven't been loaded")
        Expects(
            account.isLoaded(),
            "Could not add account because the account hasn't been loaded",
        )

        account.localKey = self.localKey

        self.__accounts.append(account)

        if self.mainAccount == None:
            self.__mainAccount = self.__accounts[0]

    def __loadFromTData(self) -> None:
        # Intended for internal usage only

        Expects(
            self.basePath != None and self.basePath != "",
            "No folder provided to load tdata",
        )

        self.accounts.clear()

        # READ KEY_DATA
        keyData = td.Storage.ReadFile("key_" + self.keyFile, self.basePath)  # type: ignore

        salt, keyEncrypted, infoEncrypted = QByteArray(), QByteArray(), QByteArray()

        keyData.stream >> salt >> keyEncrypted >> infoEncrypted

        Expects(
            keyData.stream.status() == QDataStream.Status.Ok,
            QDataStreamFailed("Failed to stream keyData"),
        )

        self.__AppVersion = keyData.version
        self.__passcodeKey = td.Storage.CreateLocalKey(
            salt, QByteArray(self.__passcodeBytes)
        )

        keyInnerData = td.Storage.DecryptLocal(keyEncrypted, self.passcodeKey)  # type: ignore

        self.__localKey = td.AuthKey(keyInnerData.stream.readRawData(256))
        self.__passcodeKeyEncrypted = keyEncrypted
        self.__passcodeKeySalt = salt

        info = td.Storage.DecryptLocal(infoEncrypted, self.localKey)  # type: ignore
        count = info.stream.readInt32()

        Expects(count > 0, "accountsCount is zero, the data might has been corrupted")

        for i in range(count):
            index = info.stream.readInt32()
            if (index >= 0) and (index < TDesktop.kMaxAccounts):
                try:
                    account = td.Account(
                        self,
                        basePath=self.basePath,
                        api=self.api,
                        keyFile=self.keyFile,
                        index=index,
                    )
                    account.prepareToStart(self.__localKey)

                    if account.isLoaded():
                        self.accounts.append(account)

                except OpenTeleException as e:
                    pass

        Expects(len(self.accounts) > 0, "No account has been loaded")

        self.__active_index = 0
        if not info.stream.atEnd():
            self.__active_index = info.stream.readInt32()

        for account in self.accounts:
            if account.index == self.__active_index:
                self.__mainAccount = account
                break

        if not self.__mainAccount:
            self.__mainAccount = self.accounts[0]

        self.__isLoaded = True

    # EXTENDED FUNCTION ====================================================================
    # @extend_classs
    # class TDesktop:

    @typing.overload
    async def ToTelethon(
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
    ) -> tl.TelegramClient:
        """
        
        ### Arguments:
            session (`str`, `Session`, default=`None`):
                The file name of the `session file` to be used, if `None` then the session will not be saved.\\
                Read more [here](https://docs.telethon.dev/en/latest/concepts/sessions.html?highlight=session#what-are-sessions).
        
            flag (`LoginFlag`, default=`CreateNewSession`):
                The login flag. Read more `[here](LoginFlag)`.
        
            api (`APIData`, default=`TelegramDesktop`):
                Which API to use. Read more `[here](API)`.
        
            password (`str`, default=`None`):
                Two-step verification password if needed.
        
        ### Returns:
            - Return an instance of `TelegramClient` on success

        ### Examples:
            Create a telethon session from tdata folder:
        ```python
            # Using the API that we've generated before. Please refer to method API.Generate() to learn more.
            oldAPI = API.TelegramDesktop.Generate(system="windows", unique_id="old_tdata")
            tdesk = TDesktop("old_tdata", api=oldAPI)

            # We can safely authorize the new client with a different API.
            newAPI = API.TelegramAndroid.Generate(unique_id="new.session")
            client = await tdesk.ToTelethon(session="new.session", flag=CreateNewSession, api=newAPI)
            await client.connect()
            await client.PrintSessions()
        ```
        """

    @typing.overload
    async def ToTelethon(
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        *,
        connection: typing.Type[Connection] = ConnectionTcpFull,
        use_ipv6: bool = False,
        proxy: Union[tuple, dict] = None,
        local_addr: Union[str, tuple] = None,
        timeout: int = 10,
        request_retries: int = 5,
        connection_retries: int = 5,
        retry_delay: int = 1,
        auto_reconnect: bool = True,
        sequential_updates: bool = False,
        flood_sleep_threshold: int = 60,
        raise_last_call_error: bool = False,
        loop: asyncio.AbstractEventLoop = None,
        base_logger: Union[str, logging.Logger] = None,
        receive_updates: bool = True
    ) -> tl.TelegramClient:
        pass

    async def ToTelethon(
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        **kwargs
    ) -> tl.TelegramClient:

        Expects(
            self.isLoaded(),
            TDesktopNotLoaded("You need to load accounts from a tdata folder first"),
        )
        Expects(
            self.accountsCount > 0,
            TDesktopHasNoAccount("There is no account in this instance of TDesktop"),
        )
        assert self.mainAccount

        return await tl.TelegramClient.FromTDesktop(
            self.mainAccount,
            session=session,
            flag=flag,
            api=api,
            password=password,
            **kwargs
        )

    @staticmethod
    async def FromTelethon(
        telethonClient: tl.TelegramClient,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
    ) -> TDesktop:
        """
        Create an instance of `TDesktop` from `TelegramClient`.

        ### Arguments:
            telethonClient (`TelegramClient`):
                The `TelegramClient` you want to convert from.

            flag (`LoginFlag`, default=`CreateNewSession`):
                The login flag. Read more `[here](LoginFlag)`.

            api (`APIData`, default=`API.TelegramDesktop`):
                Which API to use. Read more `[here](API)`.

            password (`str`, default=`None`):
                Two-step verification password if needed.

        ### Examples:
            Save a telethon session to tdata:
        ```python
            # Using the API that we've generated before. Please refer to method API.Generate() to learn more.
            oldAPI = API.TelegramDesktop.Generate(system="windows", unique_id="old.session")
            oldclient = TelegramClient("old.session", api=oldAPI)
            await oldClient.connect()

            # We can safely CreateNewSession with a different API.
            # Be aware that you should not use UseCurrentSession with a different API than the one that first authorized it.
            newAPI = API.TelegramAndroid.Generate("new_tdata")
            tdesk = await TDesktop.FromTelethon(oldclient, flag=CreateNewSession, api=newAPI)

            # Save the new session to a folder named "new_tdata"
            tdesk.SaveTData("new_tdata")
        ```
        """

        Expects(
            (flag == CreateNewSession) or (flag == UseCurrentSession),
            LoginFlagInvalid("LoginFlag invalid"),
        )

        _self = TDesktop()
        _self.__generateLocalKey()

        await td.Account.FromTelethon(
            telethonClient, flag=flag, api=api, password=password, owner=_self
        )

        return _self

    @classmethod
    def PerformanceMode(cls, enabled: bool = True):
        """
        Enable or disable performance mode. See `kPerformanceMode`.
        It is enabled by default.

        ### Arguments:
            enabled (`bool`, default=`True`):
                Either enable or disable performance mode.

        """
        cls.kPerformanceMode = enabled

    kMaxAccounts: int = int(3)
    """The maximum amount of accounts a client can have"""

    kDefaultKeyFile: str = "data"
    """See `TDesktop.keyFile`"""

    kPerformanceMode: bool = True
    """
    When enabled, `SaveTData()` will be 5000x faster.
    - What it does is using a constant `localKey` rather than generating it everytime when saving tdata.
    - The average time for generating `localKey` is about `250` to `350` ms, depend on your CPU.
    - When in performance mode, the average time to generate `localKey` is `0.0628` ms. Which is 5000x faster
    - Of course this comes with a catch, your `tdata files` will always use a same constant `localKey`. Basicly no protection at all, but who cares?
    
    ### Notes:
        Note: Performance mode will be disabled if `passcode` is set.
    """

    @property
    def api(self) -> APIData:
        """
        The API this client is using.
        """
        return self.__api

    @api.setter
    def api(self, value) -> None:
        self.__api = value
        for account in self.accounts:
            account.api = value

    @property
    def basePath(self) -> Optional[str]:
        """
        Base folder of `TDesktop`, this is where data stored
        Same as tdata folder of `Telegram Desktop`
        """
        return self.__basePath

    @property
    def passcode(self) -> str:
        """
        Passcode used to encrypt and decrypt data
        Same as the Local Passcode of `Telegram Desktop`
        """
        return self.__passcode

    @property
    def keyFile(self) -> str:
        """
        The default value is `"data"`, this argument is rarely ever used.
        It is used by `Telegram Desktop` by running it with the `"-key"` argument.
        I don't know what's the use cases of it, maybe this was a legacy feature of `Telegram Desktop`.
        """
        return self.__keyFile

    @property
    def passcodeKey(self) -> Optional[td.AuthKey]:
        return self.__passcodeKey

    @property
    def localKey(self) -> Optional[td.AuthKey]:
        """
        The key used to encrypt/decrypt data
        """
        return self.__localKey

    @property
    def AppVersion(self) -> Optional[int]:
        """
        App version of TDesktop client
        """
        return self.__AppVersion

    @property
    def AppVersionString(self) -> Optional[str]:
        raise NotImplementedError()
        return self.__AppVersion

    @property
    def accountsCount(self) -> int:
        """
        The number of accounts this client has
        """
        # return self.__accountsCount
        return len(self.__accounts)

    @property
    def accounts(self) -> List[td.Account]:
        """
        List of accounts this client has\n
        If you want to get the main account, please use .mainAccount instead
        """
        return self.__accounts

    @property
    def mainAccount(self) -> Optional[td.Account]:
        """
        The main account of the client
        """
        return self.__mainAccount



from __future__ import annotations

from . import debug

from typing import Coroutine, Tuple, Type, Callable, TypeVar, Optional, List, Any, Dict
from types import FunctionType

import abc

APP_VERSION = 3004000
TDF_MAGIC = b"TDF$"

_T = TypeVar("_T")
_TCLS = TypeVar("_TCLS", bound=type)
_RT = TypeVar("_RT")
_F = TypeVar("_F", bound=Callable[..., Any])


class BaseMetaClass(abc.ABCMeta):  # pragma: no cover
    def __new__(
        cls: Type[_T], clsName: str, bases: Tuple[type], attrs: Dict[str, Any]
    ) -> _T:

        # Hook all subclass methods
        if debug.IS_DEBUG_MODE:  # pragma: no cover
            ignore_list = [
                "__new__",
                "__del__",
                "__get__",
                "__call__",
                "__set_name__",
                "__str__",
                "__repr__",
            ]

            for attr, val in attrs.items():
                if (
                    not attr in ignore_list
                    and callable(val)
                    and not isinstance(val, type)
                ):
                    newVal = debug.DebugMethod(val)
                    attrs[attr] = newVal

        result = super().__new__(cls, clsName, bases, attrs)

        return result


class BaseObject(object, metaclass=BaseMetaClass):
    pass


class override(object):  # nocov
    """
    To use inside a class decorated with @extend_class\n
    Any attributes decorated with @override will be replaced
    """

    def __new__(cls, decorated_func: _F) -> _F:

        # check if decorated_cls really is a function
        if not isinstance(decorated_func, FunctionType):
            raise BaseException(
                "@override decorator is only for functions, not classes"
            )

        decorated_func.__isOverride__ = True  # type: ignore
        return decorated_func  # type: ignore

    @staticmethod
    def isOverride(func: _F) -> bool:
        if not hasattr(func, "__isOverride__"):
            return False
        return func.__isOverride__


class extend_class(object):  # nocov
    """
    Extend a class, all attributes will be added to its parents\n
    This won't override attributes that are already existed, please refer to @override or @extend_override_class to do this
    """

    def __new__(cls, decorated_cls: _TCLS, isOverride: bool = False) -> _TCLS:

        # check if decorated_cls really is a class (type)
        if not isinstance(cls, type):
            raise BaseException(
                "@extend_class decorator is only for classes, not functions"
            )

        newAttributes = dict(decorated_cls.__dict__)
        crossDelete = ["__abstractmethods__", "__module__", "_abc_impl", "__doc__", "__firstlineno__", "__static_attributes__"]
        [
            (newAttributes.pop(cross) if cross in newAttributes else None)
            for cross in crossDelete
        ]

        crossDelete = {}

        base = decorated_cls.__bases__[0]

        if not isOverride:
            # loop through its parents and add attributes

            for attributeName, attributeValue in newAttributes.items():

                # check if class base already has this attribute
                result = extend_class.getattr(base, attributeName)

                if result != None:
                    if id(result["value"]) == id(attributeValue):
                        crossDelete[attributeName] = attributeValue
                    else:

                        # if not override this attribute
                        if not override.isOverride(attributeValue):
                            print(
                                f"[{attributeName}] {id(result['value'])} - {id(attributeValue)}"
                            )
                            raise BaseException("err")

            [newAttributes.pop(cross) for cross in crossDelete]

        for attributeName, attributeValue in newAttributes.items():

            # let's backup this attribute for future uses
            result = extend_class.getattr(base, attributeName)

            if result != None:
                # ! dirty code, gonna fix it later, it's okay for now
                setattr(
                    base,
                    f"__{decorated_cls.__name__}__{attributeName}",
                    result["value"],
                )
                setattr(
                    decorated_cls,
                    f"__{decorated_cls.__name__}__{attributeName}",
                    result["value"],
                )

            setattr(base, attributeName, attributeValue)

        return decorated_cls

    @staticmethod
    def object_hierarchy_getattr(obj: object, attributeName: str) -> List[str]:

        results = []
        if type(obj) == object:
            return results

        if attributeName in obj.__dict__:
            val = obj.__dict__[attributeName]
            results.append({"owner": obj, "value": val})

        if attributeName in obj.__class__.__dict__:
            val = obj.__class__.__dict__[attributeName]
            results.append({"owner": obj, "value": val})

        for base in obj.__bases__:  # type: ignore
            results += extend_class.object_hierarchy_getattr(base, attributeName)

        results.reverse()
        return results

    @staticmethod
    def getattr(obj: object, attributeName: str) -> Optional[dict]:
        try:
            value = getattr(obj, attributeName)
            return {"owner": obj, "value": value}
        except BaseException as e:
            return None


class extend_override_class(extend_class):
    """
    Extend a class, all attributes will be added to its parents\n
    If those attributes are already existed, they will be replaced by the new one
    """

    def __new__(cls, decorated_cls: _TCLS) -> _TCLS:
        return super().__new__(cls, decorated_cls, True)


class sharemethod(type):
    def __get__(self, obj, cls):
        self.__owner__ = obj if obj else cls
        return self

    def __call__(self, *args) -> Any:
        return self.__fget__.__get__(self.__owner__)(*args)  # type: ignore

    def __set_name__(self, owner, name):
        self.__owner__ = owner

    def __new__(cls: Type[_T], func: _F) -> Type[_F]:

        clsName = func.__class__.__name__
        bases = func.__class__.__bases__
        attrs = func.__dict__
        # attrs = dict(func.__class__.__dict__)
        result = super().__new__(cls, clsName, bases, attrs)
        result.__fget__ = func

        return result



from __future__ import annotations

from .configs import *
from . import shared as td

from telethon.network.connection.connection import Connection
from telethon.network.connection.tcpfull import ConnectionTcpFull
from telethon.sessions.abstract import Session

import logging

# if TYPE_CHECKING:
#     from ..opentele import *


class MapData(BaseObject):  # nocov
    def __init__(self, basePath: str) -> None:

        self.basePath = basePath

        self._draftsMap: Dict[PeerId, FileKey] = {}
        self._draftCursorsMap: Dict[PeerId, FileKey] = {}
        self._draftsNotReadMap: Dict[PeerId, bool] = {}

        self._locationsKey = FileKey(0)
        self._trustedBotsKey = FileKey(0)
        self._installedStickersKey = FileKey(0)
        self._featuredStickersKey = FileKey(0)
        self._recentStickersKey = FileKey(0)
        self._favedStickersKey = FileKey(0)
        self._archivedStickersKey = FileKey(0)
        self._archivedMasksKey = FileKey(0)
        self._installedCustomEmojiKey = FileKey(0)
        self._featuredCustomEmojiKey = FileKey(0)
        self._archivedCustomEmojiKey = FileKey(0)
        self._searchSuggestionsKey = FileKey(0)
        self._webviewStorageTokenBots = FileKey(0)
        self._webviewStorageTokenOther = FileKey(0)
        self._savedGifsKey = FileKey(0)
        self._recentStickersKeyOld = FileKey(0)
        self._legacyBackgroundKeyDay = FileKey(0)
        self._legacyBackgroundKeyNight = FileKey(0)

        # self._settingsKey = FileKey(0)
        # ! TO BE ADDED
        self._settingsKey = FileKey(1851671142505648812)
        """MUST HAVE OR CAN'T WRITE MAPDATA, I'M WOKRING TO FIX THIS"""

        self._recentHashtagsAndBotsKey = FileKey(0)
        self._exportSettingsKey = FileKey(0)
        self._installedMasksKey = FileKey(0)
        self._recentMasksKey = FileKey(0)

    def read(self, localKey: td.AuthKey, legacyPasscode: QByteArray) -> None:

        try:
            mapData = td.Storage.ReadFile("map", self.basePath)
        except OpenTeleException as e:
            raise TDataReadMapDataFailed(
                "Could not read map data, find not found or couldn't be opened"
            ) from e

        legacySalt, legacyKeyEncrypted, mapEncrypted = (
            QByteArray(),
            QByteArray(),
            QByteArray(),
        )

        mapData.stream >> legacySalt >> legacyKeyEncrypted >> mapEncrypted

        ExpectStreamStatus(mapData.stream, "Could not stream data from mapData")

        if not localKey:

            # LocalEncryptSaltSize = 32
            Expects(
                legacySalt.size() == 32,
                TDataReadMapDataFailed(
                    f"Bad salt in map file, size: {legacySalt.size()}"
                ),
            )

            legacyPasscodeKey = td.Storage.CreateLegacyLocalKey(
                legacySalt, legacyPasscode
            )

            try:
                keyData = td.Storage.DecryptLocal(legacyKeyEncrypted, legacyPasscodeKey)
            except OpenTeleException as e:
                raise TDataReadMapDataIncorrectPasscode(
                    "Could not decrypt pass-protected key from map file, maybe bad password..."
                ) from e

            localKey = td.AuthKey.FromStream(keyData.stream)

        try:
            map = td.Storage.DecryptLocal(mapEncrypted, localKey)
        except OpenTeleException as e:
            raise TDataReadMapDataFailed("Could not decrypt map data") from e

        selfSerialized = QByteArray()
        draftsMap: typing.Dict[PeerId, FileKey] = {}
        draftCursorsMap: typing.Dict[PeerId, FileKey] = {}
        draftsNotReadMap: typing.Dict[PeerId, bool] = {}

        locationsKey = 0
        reportSpamStatusesKey = 0
        trustedBotsKey = 0
        recentStickersKeyOld = 0
        installedStickersKey = 0
        featuredStickersKey = 0
        recentStickersKey = 0
        favedStickersKey = 0
        archivedStickersKey = 0
        installedMasksKey = 0
        recentMasksKey = 0
        archivedMasksKey = 0
        installedCustomEmojiKey = 0
        featuredCustomEmojiKey = 0
        archivedCustomEmojiKey = 0
        searchSuggestionsKey = 0
        webviewStorageTokenBots = 0
        webviewStorageTokenOther = 0
        savedGifsKey = 0
        legacyBackgroundKeyDay = 0
        legacyBackgroundKeyNight = 0
        userSettingsKey = 0
        recentHashtagsAndBotsKey = 0
        exportSettingsKey = 0

        is_finished = False

        while not is_finished and not map.stream.atEnd():
            keyType = map.stream.readUInt32()

            if keyType == lskType.lskDraft:
                count = map.stream.readUInt32()
                for i in range(count):
                    key = FileKey(map.stream.readUInt64())
                    peerIdSerialized = map.stream.readUInt64()
                    peerId = PeerId.FromSerialized(peerIdSerialized)
                    draftsMap[peerId] = key
                    draftsNotReadMap[peerId] = True

            elif keyType == lskType.lskSelfSerialized:
                map.stream >> selfSerialized

            elif keyType == lskType.lskDraftPosition:
                count = map.stream.readUInt32()
                for i in range(count):
                    key = FileKey(map.stream.readUInt64())
                    peerIdSerialized = map.stream.readUInt64()
                    peerId = PeerId.FromSerialized(peerIdSerialized)
                    draftCursorsMap[peerId] = key

            elif (
                (keyType == lskType.lskLegacyImages)
                or (keyType == lskType.lskLegacyStickerImages)
                or (keyType == lskType.lskLegacyAudios)
            ):
                count = map.stream.readUInt32()
                for i in range(count):
                    filekey = map.stream.readUInt64()
                    first = map.stream.readUInt64()
                    second = map.stream.readUInt64()
                    size = map.stream.readInt32()

            elif keyType == lskType.lskLocations:
                locationsKey = map.stream.readUInt64()

            elif keyType == lskType.lskReportSpamStatusesOld:
                reportSpamStatusesKey = map.stream.readUInt64()
                # ClearKey(reportSpamStatusesKey, _basePath);

            elif keyType == lskType.lskTrustedBots:
                trustedBotsKey = map.stream.readUInt64()

            elif keyType == lskType.lskRecentStickersOld:
                recentStickersKeyOld = map.stream.readUInt64()

            elif keyType == lskType.lskBackgroundOldOld:
                # TO BE ADDED
                # map.stream >> (Window::Theme::IsNightMode()
                #     ? legacyBackgroundKeyNight
                #     : legacyBackgroundKeyDay);
                map.stream >> legacyBackgroundKeyDay

            elif keyType == lskType.lskBackgroundOld:
                legacyBackgroundKeyDay = map.stream.readUInt64()
                legacyBackgroundKeyNight = map.stream.readUInt64()

            elif keyType == lskType.lskUserSettings:
                userSettingsKey = map.stream.readUInt64()

            elif keyType == lskType.lskRecentHashtagsAndBots:
                recentHashtagsAndBotsKey = map.stream.readUInt64()

            elif keyType == lskType.lskStickersOld:
                installedStickersKey = map.stream.readUInt64()

            elif keyType == lskType.lskStickersKeys:
                installedStickersKey = map.stream.readUInt64()
                featuredStickersKey = map.stream.readUInt64()
                recentStickersKey = map.stream.readUInt64()
                archivedStickersKey = map.stream.readUInt64()

            elif keyType == lskType.lskFavedStickers:
                favedStickersKey = map.stream.readUInt64()

            elif keyType == lskType.lskSavedGifsOld:
                key = map.stream.readUInt64()

            elif keyType == lskType.lskSavedGifs:
                savedGifsKey = map.stream.readUInt64()

            elif keyType == lskType.lskSavedPeersOld:
                key = map.stream.readUInt64()

            elif keyType == lskType.lskExportSettings:
                exportSettingsKey = map.stream.readUInt64()

            elif keyType == lskType.lskMasksKeys:
                installedMasksKey = map.stream.readUInt64()
                recentMasksKey = map.stream.readUInt64()
                archivedMasksKey = map.stream.readUInt64()

            elif keyType == lskType.lskCustomEmojiKeys:
                installedCustomEmojiKey = map.stream.readUInt64()
                featuredCustomEmojiKey = map.stream.readUInt64()
                archivedCustomEmojiKey = map.stream.readUInt64()

            elif keyType == lskType.lskSearchSuggestions:
                searchSuggestionsKey = map.stream.readUInt64()

            elif keyType == lskType.lskWebviewTokens:
                is_finished = True

            else:
                logging.warning(f"Unknown key type in encrypted map: {keyType}")

            ExpectStreamStatus(map.stream, "Could not stream data from mapData ")

        self.__localKey = localKey

        self._draftsMap = draftsMap
        self._draftCursorsMap = draftCursorsMap
        self._draftsNotReadMap = draftsNotReadMap

        self._locationsKey = locationsKey
        self._trustedBotsKey = trustedBotsKey
        self._recentStickersKeyOld = recentStickersKeyOld
        self._installedStickersKey = installedStickersKey
        self._featuredStickersKey = featuredStickersKey
        self._recentStickersKey = recentStickersKey
        self._favedStickersKey = favedStickersKey
        self._archivedStickersKey = archivedStickersKey
        self._savedGifsKey = savedGifsKey
        self._installedMasksKey = installedMasksKey
        self._recentMasksKey = recentMasksKey
        self._archivedMasksKey = archivedMasksKey
        self._installedCustomEmojiKey = installedCustomEmojiKey
        self._featuredCustomEmojiKey = featuredCustomEmojiKey
        self._archivedCustomEmojiKey = archivedCustomEmojiKey
        self._searchSuggestionsKey = searchSuggestionsKey
        self._webviewStorageTokenBots = webviewStorageTokenBots
        self._webviewStorageTokenOther = webviewStorageTokenOther
        self._legacyBackgroundKeyDay = legacyBackgroundKeyDay
        self._legacyBackgroundKeyNight = legacyBackgroundKeyNight
        self._settingsKey = userSettingsKey
        self._recentHashtagsAndBotsKey = recentHashtagsAndBotsKey
        self._exportSettingsKey = exportSettingsKey
        self._oldMapVersion = mapData.version

    def prepareToWrite(self) -> td.Storage.EncryptedDescriptor:
        # Intended for internal usage only

        mapSize = 0

        # TO BE ADDED
        # if (!self.isEmpty()) mapSize += sizeof(uint32) + Serialize::bytearraySize(self);
        if len(self._draftsMap) > 0:
            mapSize += sizeof(uint32) * 2 + len(self._draftsMap) * sizeof(uint64) * 2
        if len(self._draftCursorsMap) > 0:
            mapSize += (
                sizeof(uint32) * 2 + len(self._draftCursorsMap) * sizeof(uint64) * 2
            )
        if self._locationsKey:
            mapSize += sizeof(uint32) + sizeof(uint64)
        if self._trustedBotsKey:
            mapSize += sizeof(uint32) + sizeof(uint64)
        if self._recentStickersKeyOld:
            mapSize += sizeof(uint32) + sizeof(uint64)
        if (
            self._installedStickersKey
            or self._featuredStickersKey
            or self._recentStickersKey
            or self._archivedStickersKey
        ):
            mapSize += sizeof(uint32) + 4 * sizeof(uint64)

        if self._favedStickersKey:
            mapSize += sizeof(uint32) + sizeof(uint64)
        if self._savedGifsKey:
            mapSize += sizeof(uint32) + sizeof(uint64)
        if self._settingsKey:
            mapSize += sizeof(uint32) + sizeof(uint64)
        if self._recentHashtagsAndBotsKey:
            mapSize += sizeof(uint32) + sizeof(uint64)
        if self._exportSettingsKey:
            mapSize += sizeof(uint32) + sizeof(uint64)
        if self._installedMasksKey or self._recentMasksKey or self._archivedMasksKey:
            mapSize += sizeof(uint32) + 3 * sizeof(uint64)
        if self._installedCustomEmojiKey or self._featuredCustomEmojiKey or self._archivedCustomEmojiKey:
            mapSize += sizeof(uint32) + 3 * sizeof(uint64)
        if self._searchSuggestionsKey:
            mapSize += sizeof(uint32) + sizeof(uint64)
        if self._webviewStorageTokenBots or self._webviewStorageTokenOther:
            mapSize += sizeof(uint32) + 2 * sizeof(uint64)

        mapData = td.Storage.EncryptedDescriptor(mapSize)
        stream = mapData.stream

        if len(self._draftsMap) > 0:
            stream.writeUInt32(lskType.lskDraft)
            stream.writeUInt32(len(self._draftsMap))
            for key, value in self._draftsMap.items():
                stream.writeUInt64(value)
                stream.writeUInt64(PeerId(key).Serialize())

        if len(self._draftCursorsMap) > 0:
            stream.writeUInt32(lskType.lskDraftPosition)
            stream.writeUInt32(len(self._draftCursorsMap))
            for key, value in self._draftCursorsMap.items():
                stream.writeUInt64(value)
                stream.writeUInt64(PeerId(key).Serialize())

        if self._locationsKey:
            stream.writeUInt32(lskType.lskLocations)
            stream.writeUInt64(self._locationsKey)

        if self._trustedBotsKey:
            stream.writeUInt32(lskType.lskTrustedBots)
            stream.writeUInt64(self._trustedBotsKey)

        if self._recentStickersKeyOld:
            stream.writeUInt32(lskType.lskRecentStickersOld)
            stream.writeUInt64(self._recentStickersKeyOld)

        if (
            self._installedStickersKey
            or self._featuredStickersKey
            or self._recentStickersKey
            or self._archivedStickersKey
        ):
            stream.writeUInt32(lskType.lskStickersKeys)
            stream.writeUInt64(self._installedStickersKey)
            stream.writeUInt64(self._featuredStickersKey)
            stream.writeUInt64(self._recentStickersKey)
            stream.writeUInt64(self._archivedStickersKey)

        if self._favedStickersKey:
            stream.writeUInt32(lskType.lskFavedStickers)
            stream.writeUInt64(self._favedStickersKey)

        if self._savedGifsKey:
            stream.writeUInt32(lskType.lskSavedGifs)
            stream.writeUInt64(self._savedGifsKey)

        if self._settingsKey:
            stream.writeUInt32(lskType.lskUserSettings)
            stream.writeUInt64(self._settingsKey)

        if self._recentHashtagsAndBotsKey:
            stream.writeUInt32(lskType.lskRecentHashtagsAndBots)
            stream.writeUInt64(self._recentHashtagsAndBotsKey)

        if self._exportSettingsKey:
            stream.writeUInt32(lskType.lskExportSettings)
            stream.writeUInt64(self._exportSettingsKey)

        if self._installedMasksKey or self._recentMasksKey or self._archivedMasksKey:
            stream.writeUInt32(lskType.lskMasksKeys)
            stream.writeUInt64(self._installedMasksKey)
            stream.writeUInt64(self._recentMasksKey)
            stream.writeUInt64(self._archivedMasksKey)

        if self._installedCustomEmojiKey or self._featuredCustomEmojiKey or self._archivedCustomEmojiKey:
            stream.writeUInt32(lskType.lskCustomEmojiKeys)
            stream.writeUInt64(self._installedCustomEmojiKey)
            stream.writeUInt64(self._featuredCustomEmojiKey)
            stream.writeUInt64(self._archivedCustomEmojiKey)
        
        if self._searchSuggestionsKey:
            stream.writeUInt32(lskType.lskSearchSuggestions)
            stream.writeUInt64(self._searchSuggestionsKey)

        if self._webviewStorageTokenBots or self._webviewStorageTokenOther:
            stream.writeUInt32(lskType.lskWebviewTokens)
            stream.writeUInt64(self._webviewStorageTokenBots)
            stream.writeUInt64(self._webviewStorageTokenOther)

        return mapData


class StorageAccount(BaseObject):  # nocov
    """
    Storage account for reading and writing to tdata
    """

    def __init__(self, owner: Account, basePath: str, keyFile: str) -> None:
        """
        Create an instance of StorageAccount

        ### Arguments
            1. owner (Account):\n
                owner

            2. basePath (str):\n
                basePath

            3. dataName (str):\n
                dataName
        """

        self.__owner = owner
        self.__keyFile = keyFile
        self.__dataNameKey = td.Storage.ComputeDataNameKey(self.__keyFile)
        self.__baseGlobalPath = td.Storage.GetAbsolutePath(basePath)
        self.__basePath = td.Storage.PathJoin(
            self.__baseGlobalPath, td.Storage.ToFilePart(self.__dataNameKey)
        )
        self.__localKey = None

        self.__mapData = MapData(self.basePath)
        self.__config = td.MTP.Config(td.MTP.Environment.Production)
        # self.mtpData = MtpData()

    @property
    def owner(self) -> Account:
        """
        The TDesktop client that own this account
        """
        return self.__owner

    @property
    def localKey(self) -> Optional[td.AuthKey]:
        """
        The key use to encrypt/decrypt data
        """
        return self.__localKey

    @localKey.setter
    def localKey(self, value):
        # localKey setter is intended for internal usage
        self.__localKey = value

    @property
    def keyFile(self) -> str:
        return self.__keyFile

    @keyFile.setter
    def keyFile(self, value):
        self.__keyFile = value
        self.__dataNameKey = td.Storage.ComputeDataNameKey(self.__keyFile)
        self.baseGlobalPath = self.baseGlobalPath

    @property
    def baseGlobalPath(self) -> str:
        return self.__baseGlobalPath

    @baseGlobalPath.setter
    def baseGlobalPath(self, basePath):
        self.__baseGlobalPath = td.Storage.GetAbsolutePath(basePath)
        self.__basePath = td.Storage.PathJoin(
            self.__baseGlobalPath, td.Storage.ToFilePart(self.__dataNameKey)
        )

    @property
    def basePath(self) -> str:
        return self.__basePath

    @property
    def config(self) -> td.MTP.Config:
        return self.__config

    @property
    def mapData(self) -> MapData:
        return self.__mapData

    def start(self, localKey: td.AuthKey) -> td.MTP.Config:
        # Intended for internal usage only

        self.__localKey = localKey
        self.readMapWith(localKey)

        return (
            self.__config if self.owner.owner.kPerformanceMode else self.readMtpConfig()
        )

    def readMtpData(self):
        # Intended for internal usage only

        # mtp = ReadEncryptedFile(ToFilePart(self.__dataNameKey), self.__basePath, self.localKey)
        mtp = td.Storage.ReadEncryptedFile(td.Storage.ToFilePart(self.__dataNameKey), self.__baseGlobalPath, self.localKey)  # type: ignore

        blockId = mtp.stream.readInt32()

        Expects(blockId == 75, TDataInvalidMagic("Not supported file version"))

        serialized = QByteArray()
        mtp.stream >> serialized
        self.owner._setMtpAuthorization(serialized)

    def readMtpConfig(self) -> td.MTP.Config:
        # Intended for internal usage only
        Expects(
            self.localKey != None,
            AccountAuthKeyNotFound("The localKey has not been initialized yet"),
        )

        try:
            file = td.Storage.ReadEncryptedFile("config", self.basePath, self.localKey)  # type: ignore
            serialized = QByteArray()
            file.stream >> serialized

            ExpectStreamStatus(file.stream, "Could not stream data from MtpConfig")

            self.__config = td.MTP.Config.FromSerialized(serialized)
            return self.__config
        except OpenTeleException as e:
            pass

        return td.MTP.Config(td.MTP.Environment.Production)

    def readMapWith(
        self, localKey: td.AuthKey, legacyPasscode: QByteArray = QByteArray()
    ):
        # Intended for internal usage only
        try:
            self.__mapData.read(localKey, legacyPasscode)
        except OpenTeleException:
            return False

        self.readMtpData()

    def writeMtpConfig(self, basePath: str) -> None:
        # Intended for internal usage only

        Expects(
            self.localKey != None,
            "localKey not found, have you initialized me correctly?",
        )
        Expects(basePath != None and basePath != "", "basePath can't be empty")

        serialized = self.owner.MtpConfig.Serialize()
        size = td.Serialize.bytearraySize(serialized)
        file = td.Storage.FileWriteDescriptor("config", basePath)
        data = td.Storage.EncryptedDescriptor(size)
        data.stream << serialized
        file.writeEncrypted(data, self.localKey)  # type: ignore
        file.finish()

    def writeMap(self, basePath: str) -> None:
        # Intended for internal usage only

        Expects(
            self.localKey != None,
            "localKey not found, have you initialized me correctly?",
        )
        Expects(basePath != None and basePath != "", "basePath can't be empty")

        map = td.Storage.FileWriteDescriptor("map", basePath)

        # i don't know what's the purpose of this, but it's in tdesktop source code
        map.writeData(QByteArray())
        map.writeData(QByteArray())

        mapDataEncrypted = self.mapData.prepareToWrite()
        map.writeEncrypted(mapDataEncrypted, self.localKey)  # type: ignore
        map.finish()

    def writeMtpData(self, baseGlobalPath: str, dataNameKey: int) -> None:
        # Intended for internal usage only

        Expects(
            self.localKey != None,
            "localKey not found, have you initialized me correctly?",
        )
        Expects(
            baseGlobalPath != None and baseGlobalPath != "",
            "baseGlobalPath can't be empty",
        )

        serialized = self.owner.serializeMtpAuthorization()
        size = sizeof(uint32) + td.Serialize.bytearraySize(serialized)
        mtp = td.Storage.FileWriteDescriptor(
            td.Storage.ToFilePart(dataNameKey), baseGlobalPath
        )
        data = td.Storage.EncryptedDescriptor(size)
        data.stream.writeInt32(dbi.MtpAuthorization)
        data.stream << serialized
        mtp.writeEncrypted(data, self.localKey)  # type: ignore
        mtp.finish()

    def _writeData(self, baseGlobalPath: str, keyFile: str = None) -> None:
        # Intended for internal usage only

        Expects(
            baseGlobalPath != None and baseGlobalPath != "",
            "baseGlobalPath can't be empty",
        )

        if keyFile != None and self.keyFile != keyFile:
            self.__keyFile = keyFile
            dataNameKey = td.Storage.ComputeDataNameKey(self.__keyFile)
        else:
            dataNameKey = self.__dataNameKey

        basePath = td.Storage.PathJoin(
            baseGlobalPath, td.Storage.ToFilePart(dataNameKey)
        )
        self.writeMap(basePath)

        if not self.owner.owner.kPerformanceMode:
            self.writeMtpConfig(basePath)

        self.writeMtpData(baseGlobalPath, dataNameKey)


class Account(BaseObject):
    """
    Telegram Desktop account

    ### Attributes:
        api (`API`):
            The API this acount is using.

        authKey (`AuthKey`):
            The authorization key used to authorize this acocunt.

        UserId (`int`):
            User ID of this account.

        MainDcId (`DcId`):
            The main Data Center ID this account connects to.

        basePath (`str`):
            The folder where tdata is stored.

        localKey (`AuthKey`):
            Key used to encrypt and decrypt tdata.

        owner (`TDesktop`):
            `td.TDesktop` client owner of this account.

        keyFile (`str`):
            See `td.TDesktop.keyFile`.

    """

    kWideIdsTag: int = int(~0)

    def __init__(
        self,
        owner: td.TDesktop,
        basePath: str = None,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        keyFile: str = None,
        index: int = 0,
    ) -> None:
        """
        Initialized a `TDesktop` account.

        You should use `TDesktop()` or `TDesktop.FromTelethon()` instead.
        Manually using `Account()` is not recommended. But this is here for your need anyway.

        ### Arguments:
            owner (`TDesktop`):
                `TDesktop` client owner of this account.

            basePath (`str`, default=None):
                The folder where `tdata` is stored.

            api (`API`, default=`TelegramDesktop`):
                Which API to use. Read more `[here](API)`.

            keyFile (`str`, default=None):
                See `TDesktop.keyFile`.

            index (`int`, default=0):
                Index of this account in the `TDesktop` client.

        ### Notes:
            TODO: `prepareToStart()` must be call after initalizing the object.

        """
        self.__owner = owner
        self.__localKey = None
        self.__authKey = None
        self.__telethonClient = None
        self.__isLoaded = False
        self.__isAuthorized = False
        self.__UserId = 0
        self.__MainDcId = DcId(0)

        self.__basePath = td.Storage.GetAbsolutePath(basePath)
        self.__keyFile = keyFile if (keyFile != None) else td.TDesktop.kDefaultKeyFile

        self.__mtpKeys: typing.List[td.AuthKey] = []
        self.__mtpKeysToDestroy: typing.List[td.AuthKey] = []
        self.api = api

        self._local = StorageAccount(
            self, self.basePath, td.Storage.ComposeDataString(self.__keyFile, index)
        )
        self.index = index

    @property
    def api(self) -> APIData:
        """
        The API this acount is using.
        """
        return self.__api

    @api.setter
    def api(self, value) -> None:
        self.__api = value
        if self.owner.api != self.api:
            self.owner.api = self.api

    @property
    def owner(self) -> td.TDesktop:
        """
        TDesktop client owner of this account.
        """
        return self.__owner

    @property
    def basePath(self) -> str:
        """
        The folder where tdata is stored.
        """
        return self.__basePath

    @property
    def keyFile(self) -> str:
        """
        See `TDesktop.keyFile`
        """
        return self.__keyFile

    @keyFile.setter
    def keyFile(self, value):
        self.__keyFile = value
        self._local.keyFile = td.Storage.ComposeDataString(value, self.index)

    @property
    def localKey(self) -> Optional[td.AuthKey]:
        """
        Key used to encrypt and decrypt tdata.
        """
        return self.__localKey

    @localKey.setter
    def localKey(self, value):
        self.__localKey = value
        self._local.localKey = value

    @property
    def authKey(self) -> Optional[td.AuthKey]:
        """
        The authorization key used to authorize this acocunt.
        """
        return self.__authKey

    @property
    def UserId(self) -> int:
        """
        User ID of this account.
        """
        return self.__UserId

    @property
    def MainDcId(self) -> DcId:
        """
        The main Data Center ID this account connects to.
        """
        return self.__MainDcId

    @property
    def MtpConfig(self) -> td.MTP.Config:
        return self._local.config

    @property
    def MapData(self) -> MapData:
        return self._local.mapData

    def isAuthorized(self) -> bool:
        return self.__isAuthorized

    def isLoaded(self) -> bool:
        return self.__isLoaded

    def start(self) -> bool:
        raise NotImplementedError()
        return self.isAuthorized()

    def prepareToStart(self, localKey: td.AuthKey) -> td.MTP.Config:
        """
        Prepare the account before starting it

        ### Arguments:
            localKey (`AuthKey`):
                `API`

        ### Returns:
            `MTP.Config`: [description]
        """

        self.__localKey = localKey
        self.__MtpConfig = self._local.start(localKey)
        return self.__MtpConfig

    def _setMtpAuthorizationCustom(
        self,
        dcId: DcId,
        userId: int,
        mtpKeys: List[td.AuthKey],
        mtpKeysToDestroy: List[td.AuthKey] = [],
    ):
        # Intended for internal usage only

        self.__MainDcId = dcId
        self.__UserId = userId
        self.__mtpKeys = mtpKeys

        for key in self.__mtpKeys:
            if key.dcId == self.MainDcId:
                self.__authKey = key
                break

        Expects(
            self.authKey != None,
            exception=TDataAuthKeyNotFound(
                "Could not find the main authKey, are you sure the data is correct?"
            ),
        )

        self.__isLoaded = True

    def _setMtpAuthorization(self, serialized: QByteArray):
        # Intended for internal usage only
        stream = QDataStream(serialized)
        stream.setVersion(QDataStream.Version.Qt_5_1)

        self.__UserId = stream.readInt32()
        self.__MainDcId = DcId(stream.readInt32())

        if ((self.__UserId << 32) | self.__MainDcId) == Account.kWideIdsTag:
            self.__UserId = stream.readUInt64()
            self.__MainDcId = DcId(stream.readInt32())

        Expects(
            stream.status() == QDataStream.Status.Ok,
            QDataStreamFailed("Could not read main fields from mtp authorization."),
        )

        def readKeys(keys: typing.List[td.AuthKey]):

            key_count = stream.readInt32()
            Expects(
                stream.status() == QDataStream.Status.Ok,
                QDataStreamFailed("Could not read keys count from mtp authorization."),
            )

            for i in range(key_count):
                dcId = DcId(stream.readInt32())
                keys.append(
                    td.AuthKey.FromStream(stream, td.AuthKeyType.ReadFromFile, dcId)
                )

        self.__mtpKeys.clear()
        self.__mtpKeysToDestroy.clear()

        readKeys(self.__mtpKeys)
        readKeys(self.__mtpKeysToDestroy)

        for key in self.__mtpKeys:
            if key.dcId == self.MainDcId:
                self.__authKey = key
                break

        Expects(
            self.__authKey != None,
            exception=TDataAuthKeyNotFound(
                "Could not find authKey, the data might has been corrupted"
            ),
        )

        self.__isLoaded = True
        # TO BE ADDED
        # while (not mtp.stream.atEnd()):
        #     blockId = mtp.stream.readInt32()
        #     if mtp.stream.status() != 0:
        #         raise OpenTeleException(OpenTeleErrorCode.QDataStreamFailed, "Failed to stream mtpData")
        #     print(blockId)

    def serializeMtpAuthorization(self) -> QByteArray:
        # Intended for internal usage only
        Expects(self.isLoaded(), "Account data not loaded yet")

        def keysSize(list: typing.List[td.AuthKey]):
            return 4 + len(list) * (4 + td.AuthKey.kSize)

        def writeKeys(stream: QDataStream, keys: typing.List[td.AuthKey]):

            stream.writeInt32(len(keys))
            for key in keys:
                stream.writeInt32(key.dcId)
                stream.writeRawData(key.key)

        result = QByteArray()
        stream = QDataStream(result, QIODevice.OpenModeFlag.WriteOnly)
        stream.setVersion(QDataStream.Version.Qt_5_1)

        stream.writeInt64(Account.kWideIdsTag)
        stream.writeInt64(self.UserId)
        stream.writeInt32(self.MainDcId)

        writeKeys(stream, self.__mtpKeys)
        writeKeys(stream, self.__mtpKeysToDestroy)
        return result

    def _writeData(self, baseGlobalPath: str, keyFile: str = None) -> None:
        # Intended for internal usage only

        self._local._writeData(baseGlobalPath, keyFile)

    def SaveTData(
        self, basePath: str = None, passcode: str = None, keyFile: str = None
    ) -> None:
        """
        Save this account to a folder

        ### Arguments:
            basePath (`str`, default=`None`):
                The path to the folder. Defaults to None.

            passcode (`str`, default=`None`):
                Lock the data with a passcode. Defaults to None.

            keyFile (`str`, default=`None`):
                See `TDesktop.keyFile`

        ### Examples
            Add an account to `TDesktop` and save it to `tdata`:
        ```python
            telethonClient = TelegramClient("sessionFile", API_ID, API_HASH)
            td = TDesktop("new_tdata")
            account = Account.FromTelethon(telethonClient, owner=td) # add this account to td
            td.SaveTData()
        ```
        """

        if basePath == None:
            basePath = self.basePath

        basePath = td.Storage.GetAbsolutePath(basePath)

        if self.basePath == None:
            self.__basePath = basePath

        self.owner.SaveTData(basePath, passcode, keyFile)

    # @extend_class
    # class Account(Account):

    @typing.overload
    async def ToTelethon(
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
    ) -> tl.TelegramClient:
        pass

    @typing.overload
    async def ToTelethon(
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        *,
        connection: typing.Type[Connection] = ConnectionTcpFull,
        use_ipv6: bool = False,
        proxy: Union[tuple, dict] = None,
        local_addr: Union[str, tuple] = None,
        timeout: int = 10,
        request_retries: int = 5,
        connection_retries: int = 5,
        retry_delay: int = 1,
        auto_reconnect: bool = True,
        sequential_updates: bool = False,
        flood_sleep_threshold: int = 60,
        raise_last_call_error: bool = False,
        loop: asyncio.AbstractEventLoop = None,
        base_logger: Union[str, logging.Logger] = None,
        receive_updates: bool = True,
    ) -> tl.TelegramClient:
        pass

    async def ToTelethon(
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        **kwargs,
    ) -> tl.TelegramClient:

        Expects(
            self.isLoaded(),
            TDAccountNotLoaded(
                "I'm not loaded yet, are you sure you're using me correctly?"
            ),
        )

        return await tl.TelegramClient.FromTDesktop(
            self, session=session, flag=flag, api=api, password=password, **kwargs
        )

    @staticmethod
    async def FromTelethon(
        telethonClient: tl.TelegramClient,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        owner: td.TDesktop = None,
    ):

        Expects(
            (flag == CreateNewSession) or (flag == UseCurrentSession),
            LoginFlagInvalid("LoginFlag invalid"),
        )

        if flag == CreateNewSession:
            if not telethonClient.is_connected():
                try:
                    await telethonClient.connect()
                except OSError as e:
                    raise TelethonUnauthorized(
                        "Could not connect telethon client to the server, please try to connect manually"
                    ) from e

                Expects(
                    await telethonClient.is_user_authorized(),
                    exception=TelethonUnauthorized(
                        "Telethon client is unauthorized, it need to be authorized first"
                    ),
                )

            copy = await telethonClient.QRLoginToNewClient(api=api, password=password)
            await copy.get_me()
        else:
            copy = telethonClient

        ss = copy.session
        authKey = ss.auth_key.key
        dcId = DcId(ss.dc_id)
        userId = copy.UserId
        authKey = td.AuthKey(authKey, td.AuthKeyType.ReadFromFile, dcId)
        
        if userId == None:
            await copy.connect()
            await copy.get_me()
            userId = copy.UserId
            
        newAccount = None

        if owner != None:

            Expects(
                owner.accountsCount < td.TDesktop.kMaxAccounts,
                exception=MaxAccountLimit(
                    "You can't have more than 3 accounts in one TDesktop clent.\n"
                    "Please create another instance of TDesktop or use Account.FromTelethon() to create an Account() independently"
                ),
            )

            index = owner.accountsCount
            newAccount = Account(
                owner=owner,
                basePath=owner.basePath,
                api=api,
                keyFile=owner.keyFile,
                index=index,
            )
            newAccount._setMtpAuthorizationCustom(dcId, userId, [authKey])  # type: ignore
            owner._addSingleAccount(newAccount)

        else:
            index = 0
            newOwner = td.TDesktop()
            newAccount = Account(owner=newOwner, api=api, index=index)
            newAccount._setMtpAuthorizationCustom(dcId, userId, [authKey])  # type: ignore
            newOwner._addSingleAccount(newAccount)

        return newAccount



from ..exception import *
from ..utils import *
from ..api import APIData, API, LoginFlag, CreateNewSession, UseCurrentSession
from .. import td

from typing import Union, Callable, TypeVar, Type, List, Dict, Any, TYPE_CHECKING
from ctypes import (
    sizeof,
    c_int32 as int32,
    c_int64 as int64,
    c_uint32 as uint32,
    c_uint64 as uint64,
    c_short as short,
    c_ushort as ushort,
)

import telethon
from telethon.sessions import StringSession
from telethon.crypto import AuthKey
from telethon import tl, functions, types, utils, password as pwd_mod

from telethon.network.connection.connection import Connection
from telethon.network.connection.tcpfull import ConnectionTcpFull


from telethon.sessions.abstract import Session
from telethon.sessions.sqlite import SQLiteSession
from telethon.sessions.memory import MemorySession

import asyncio



import typing


from .account import Account, StorageAccount, MapData
from .auth import AuthKey, AuthKeyType
from .mtp import MTP
from .storage import Storage, Serialize
from .tdesktop import TDesktop
from . import configs

from .. import exception as excpt

from ..api import APIData, API

from typing import Optional

import struct
import os



from __future__ import annotations

from enum import IntEnum

from ..exception import *
from ..utils import *
from ..api import *
from .. import tl

from typing import (
    Union,
    Callable,
    TypeVar,
    Type,
    Optional,
    List,
    Dict,
    Any,
    TYPE_CHECKING,
)
from ctypes import (
    sizeof,
    c_int32 as int32,
    c_int64 as int64,
    c_uint32 as uint32,
    c_uint64 as uint64,
    c_short as short,
    c_ushort as ushort,
)
from PyQt5.QtCore import (
    QByteArray,
    QDataStream,
    QBuffer,
    QIODevice,
    QSysInfo,
    QDir,
    QFile,
)
from types import FunctionType

import asyncio


APP_VERSION = 3004000
TDF_MAGIC = b"TDF$"


_T = TypeVar("_T")
_T2 = TypeVar("_T2")
_TCLS = TypeVar("_TCLS", bound=type)
_RT = TypeVar("_RT")

_F = TypeVar("_F", bound=Callable[..., Any])


class BareId(int):  # nocov
    """
    BareId
    """


class ChatIdType(BaseObject):  # nocov
    """
    ChatIdType
    """

    bare = BareId(0)
    kShift = BareId(0)
    kReservedBit = BareId(0x80)

    def __init__(self, value: BareId) -> None:
        self.bare = value


class UserId(ChatIdType):  # nocov
    kShift = BareId(0)


class ChatId(ChatIdType):  # nocov
    kShift = BareId(1)


class ChannelId(ChatIdType):  # nocov
    kShift = BareId(2)


class FakeChatId(ChatIdType):  # nocov
    kShift = BareId(0x7F)


class PeerId(int):  # nocov
    """
    PeerId
    """

    kChatTypeMask = BareId(0xFFFFFFFFFFFF)

    def __init__(self, value) -> None:
        self.value = value

    def Serialize(self):
        Expects(not (self.value & (UserId.kReservedBit << 48)))
        return self.value | (UserId.kReservedBit << 48)

    @staticmethod
    def FromChatIdType(
        id: typing.Union[UserId, ChatId, ChannelId, FakeChatId]
    ) -> PeerId:
        return PeerId(id.bare | (BareId(id.kShift) << 48))

    @staticmethod
    def FromSerialized(serialized: int) -> PeerId:
        flag = UserId.kReservedBit << 48
        legacy = not (serialized & (UserId.kReservedBit << 48))

        if not legacy:
            return PeerId(serialized & (~flag))

        PeerIdMask = int(0xFFFFFFFF)
        PeerIdTypeMask = 0xF00000000
        PeerIdUserShift = 0x000000000
        PeerIdChatShift = 0x100000000
        PeerIdChannelShift = 0x200000000
        PeerIdFakeShift = 0xF00000000

        if (serialized & PeerIdTypeMask) == PeerIdUserShift:
            return PeerId.FromChatIdType(UserId(BareId(serialized & PeerIdMask)))

        elif (serialized & PeerIdTypeMask) == PeerIdChatShift:
            return PeerId.FromChatIdType(ChatId(BareId(serialized & PeerIdMask)))

        elif (serialized & PeerIdTypeMask) == PeerIdChannelShift:
            return PeerId.FromChatIdType(ChannelId(BareId(serialized & PeerIdMask)))

        elif (serialized & PeerIdTypeMask) == PeerIdFakeShift:
            return PeerId.FromChatIdType(FakeChatId(BareId(serialized & PeerIdMask)))

        return PeerId(0)


class FileKey(int):  # nocov
    pass


class DcId(int):  # nocov
    """
    Data Center ID
    """

    kDcShift: DcId = 10000  # type: ignore
    Invalid: DcId = 0  # type: ignore
    _0: DcId = 0  # type: ignore
    _1: DcId = 1  # type: ignore
    _2: DcId = 2  # type: ignore
    _3: DcId = 3  # type: ignore
    _4: DcId = 4  # type: ignore
    _5: DcId = 5  # type: ignore

    @staticmethod
    def BareDcId(shiftedDcId: Union[ShiftedDcId, DcId]) -> DcId:
        return DcId(shiftedDcId % DcId.kDcShift)


class ShiftedDcId(DcId):  # nocov
    """
    Shifted Data Center ID
    """

    @staticmethod
    def ShiftDcId(dcId: DcId, value: int) -> ShiftedDcId:
        return ShiftedDcId(dcId + DcId.kDcShift * value)


class BuiltInDc(BaseObject):  # type: ignore # nocov
    """
    Default DC that is hard-coded
    """

    def __init__(self, id: DcId, ip: str, port: int):
        self.id = id
        self.ip = ip
        self.port = port


class BuiltInDc(BuiltInDc):  # nocov
    kBuiltInDcs = [
        BuiltInDc(DcId._1, "149.154.175.50", 443),
        BuiltInDc(DcId._2, "149.154.167.51", 443),
        BuiltInDc(DcId._2, "95.161.76.100", 443),
        BuiltInDc(DcId._3, "149.154.175.100", 443),
        BuiltInDc(DcId._4, "149.154.167.91", 443),
        BuiltInDc(DcId._5, "149.154.171.5", 443),
    ]

    kBuiltInDcsIPv6 = [
        BuiltInDc(DcId._1, "2001:0b28:f23d:f001:0000:0000:0000:000a", 443),
        BuiltInDc(DcId._2, "2001:067c:04e8:f002:0000:0000:0000:000a", 443),
        BuiltInDc(DcId._3, "2001:0b28:f23d:f003:0000:0000:0000:000a", 443),
        BuiltInDc(DcId._4, "2001:067c:04e8:f004:0000:0000:0000:000a", 443),
        BuiltInDc(DcId._5, "2001:0b28:f23f:f005:0000:0000:0000:000a", 443),
    ]

    kBuiltInDcsTest = [
        BuiltInDc(DcId._1, "149.154.175.10", 443),
        BuiltInDc(DcId._2, "149.154.167.40", 443),
        BuiltInDc(DcId._3, "149.154.175.117", 443),
    ]

    kBuiltInDcsIPv6Test = [
        BuiltInDc(DcId._1, "2001:0b28:f23d:f001:0000:0000:0000:000e", 443),
        BuiltInDc(DcId._2, "2001:067c:04e8:f002:0000:0000:0000:000e", 443),
        BuiltInDc(DcId._3, "2001:0b28:f23d:f003:0000:0000:0000:000e", 443),
    ]


class dbi(int):  # nocov
    Key = 0x00
    User = 0x01
    DcOptionOldOld = 0x02
    ChatSizeMaxOld = 0x03
    MutePeerOld = 0x04
    SendKeyOld = 0x05
    AutoStart = 0x06
    StartMinimized = 0x07
    SoundFlashBounceNotifyOld = 0x08
    WorkModeOld = 0x09
    SeenTrayTooltip = 0x0A
    DesktopNotifyOld = 0x0B
    AutoUpdate = 0x0C
    LastUpdateCheck = 0x0D
    WindowPositionOld = 0x0E
    ConnectionTypeOldOld = 0x0F
    # 0x10 reserved
    DefaultAttach = 0x11
    CatsAndDogsOld = 0x12
    ReplaceEmojiOld = 0x13
    AskDownloadPathOld = 0x14
    DownloadPathOldOld = 0x15
    ScaleOld = 0x16
    EmojiTabOld = 0x17
    RecentEmojiOldOldOld = 0x18
    LoggedPhoneNumberOld = 0x19
    MutedPeersOld = 0x1A
    # 0x1b reserved
    NotifyViewOld = 0x1C
    SendToMenu = 0x1D
    CompressPastedImageOld = 0x1E
    LangOld = 0x1F
    LangFileOld = 0x20
    TileBackgroundOld = 0x21
    AutoLockOld = 0x22
    DialogLastPath = 0x23
    RecentEmojiOldOld = 0x24
    EmojiVariantsOldOld = 0x25
    RecentStickers = 0x26
    DcOptionOld = 0x27
    TryIPv6Old = 0x28
    SongVolumeOld = 0x29
    WindowsNotificationsOld = 0x30
    IncludeMutedOld = 0x31
    MegagroupSizeMaxOld = 0x32
    DownloadPathOld = 0x33
    AutoDownloadOld = 0x34
    SavedGifsLimitOld = 0x35
    ShowingSavedGifsOld = 0x36
    AutoPlayOld = 0x37
    AdaptiveForWideOld = 0x38
    HiddenPinnedMessagesOld = 0x39
    RecentEmojiOld = 0x3A
    EmojiVariantsOld = 0x3B
    DialogsModeOld = 0x40
    ModerateModeOld = 0x41
    VideoVolumeOld = 0x42
    StickersRecentLimitOld = 0x43
    NativeNotificationsOld = 0x44
    NotificationsCountOld = 0x45
    NotificationsCornerOld = 0x46
    ThemeKeyOld = 0x47
    DialogsWidthRatioOld = 0x48
    UseExternalVideoPlayer = 0x49
    DcOptionsOld = 0x4A
    MtpAuthorization = 0x4B
    LastSeenWarningSeenOld = 0x4C
    SessionSettings = 0x4D
    LangPackKey = 0x4E
    ConnectionTypeOld = 0x4F
    StickersFavedLimitOld = 0x50
    SuggestStickersByEmojiOld = 0x51
    SuggestEmojiOld = 0x52
    TxtDomainStringOldOld = 0x53
    ThemeKey = 0x54
    TileBackground = 0x55
    CacheSettingsOld = 0x56
    AnimationsDisabled = 0x57
    ScalePercent = 0x58
    PlaybackSpeedOld = 0x59
    LanguagesKey = 0x5A
    CallSettingsOld = 0x5B
    CacheSettings = 0x5C
    TxtDomainStringOld = 0x5D
    ApplicationSettings = 0x5E
    DialogsFiltersOld = 0x5F
    FallbackProductionConfig = 0x60
    BackgroundKey = 0x61

    EncryptedWithSalt = 333
    Encrypted = 444

    Version = 666


class lskType(int):  # nocov
    lskUserMap = 0x00
    lskDraft = 0x01  # data: PeerId peer
    lskDraftPosition = 0x02  # data: PeerId peer
    lskLegacyImages = 0x03  # legacy
    lskLocations = 0x04  # no data
    lskLegacyStickerImages = 0x05  # legacy
    lskLegacyAudios = 0x06  # legacy
    lskRecentStickersOld = 0x07  # no data
    lskBackgroundOldOld = 0x08  # no data
    lskUserSettings = 0x09  # no data
    lskRecentHashtagsAndBots = 0x0A  # no data
    lskStickersOld = 0x0B  # no data
    lskSavedPeersOld = 0x0C  # no data
    lskReportSpamStatusesOld = 0x0D  # no data
    lskSavedGifsOld = 0x0E  # no data
    lskSavedGifs = 0x0F  # no data
    lskStickersKeys = 0x10  # no data
    lskTrustedBots = 0x11  # no data
    lskFavedStickers = 0x12  # no data
    lskExportSettings = 0x13  # no data
    lskBackgroundOld = 0x14  # no data
    lskSelfSerialized = 0x15  # serialized self
    lskMasksKeys = 0x16  # no data
    lskCustomEmojiKeys = 0x17  # no data
    lskSearchSuggestions = 0x18  # no data
    lskWebviewTokens = 0x19  # data: QByteArray bots, QByteArray other


class BotTrustFlag(int):  # nocov
    NoOpenGame = 1 << 0
    Payment = 1 << 1


# class QByteArray(QByteArray):
#     def dump(self):

#         print(f"  ---- hexdump ----: [0x{hex(self.size())}:{self.size()}] ({id(self)})")

#         size = self.size()
#         data = self.data()
#         out = ""
#         asc = ""

#         for i in range(size):

#             if i % 16 == 0:
#                 if i != 0:
#                     out += " |  "
#                     out += asc + "\n"
#                     asc = ""

#                 out += "  %0.4x " % i

#             out += " %0.2x" % data[i]

#             if (i > 0) and ((i % 8) == 7) and ((i % 16) != 15):
#                 out += " "

#             if (data[i] < 0x20) or (data[i] > 0x7E):
#                 asc += "."
#             else:
#                 asc += chr(data[i])

#         i = size - 1
#         if ((i % 16) != 0) and ((i % 16) == (i % 8)):
#             out += " "

#         while (i % 16) != 0:
#             out += "   "
#             i += 1

#         out += " |  "
#         print(out)


# def hexdump(byte: bytes):

#     if isinstance(byte, QByteArray):
#         byte = byte.data()

#     print(
#         f"  ---- hexdump ----: [0x{hex(byte.__len__())}:{byte.__len__()}] ({id(byte)})"
#     )

#     size = byte.__len__()
#     data = byte
#     out = ""
#     asc = ""

#     for i in range(size):

#         if i % 16 == 0:
#             if i != 0:
#                 out += " |  "
#                 out += asc + "\n"
#                 asc = ""

#             out += "  %0.4x " % i

#         out += " %0.2x" % data[i]

#         if (i > 0) and ((i % 8) == 7) and ((i % 16) != 15):
#             out += " "

#         if (data[i] < 0x20) or (data[i] > 0x7E):
#             asc += "."
#         else:
#             asc += chr(data[i])

#     i = size - 1
#     if ((i % 16) != 0) and ((i % 16) == (i % 8)):
#         out += " "

#     while (i % 16) != 0:
#         out += "   "
#         i += 1

#     out += " |  "
#     print(out)



from __future__ import annotations
from .configs import *
from . import shared as td

import hashlib
import os
import tgcrypto

# if TYPE_CHECKING:
#     from ..opentele import *


class Serialize(BaseObject):
    @staticmethod
    def bytearraySize(arr: QByteArray):
        return sizeof(uint32) + arr.size()

    @staticmethod
    def bytesSize(arr: bytes):
        return sizeof(uint32) + len(arr)

    @staticmethod
    def stringSize(arr: str):
        return sizeof(uint32) + len(arr) + sizeof(ushort)


class Storage(BaseObject):
    class ReadSettingsContext(BaseObject):  # pragma: no cover
        def __init__(self) -> None:

            self.fallbackConfigLegacyDcOptions: td.MTP.DcOptions = td.MTP.DcOptions(
                td.MTP.Environment.Production
            )
            self.fallbackConfigLegacyChatSizeMax = 0
            self.fallbackConfigLegacySavedGifsLimit = 0
            self.fallbackConfigLegacyStickersRecentLimit = 0
            self.fallbackConfigLegacyStickersFavedLimit = 0
            self.fallbackConfigLegacyMegagroupSizeMax = 0
            self.fallbackConfigLegacyTxtDomainString = str()
            self.fallbackConfig = QByteArray()

            self.cacheTotalSizeLimit = 0
            self.cacheTotalTimeLimit = 0
            self.cacheBigFileTotalSizeLimit = 0
            self.cacheBigFileTotalTimeLimit = 0

            self.themeKeyLegacy = FileKey(0)
            self.themeKeyDay = FileKey(0)
            self.themeKeyNight = FileKey(0)
            self.backgroundKeyDay = FileKey(0)
            self.backgroundKeyNight = FileKey(0)

            self.backgroundKeysRead = False
            self.tileDay = False
            self.tileNight = True
            self.tileRead = False

            self.langPackKey = FileKey(0)
            self.languagesKey = FileKey(0)

            self.mtpAuthorization = QByteArray()
            self.mtpLegacyKeys: typing.List[td.AuthKey] = []

            self.mtpLegacyMainDcId = 0
            self.mtpLegacyUserId = 0
            self.legacyRead = False

    class FileReadDescriptor(BaseObject):
        def __init__(self) -> None:
            self.__version = 0
            self.__data = QByteArray()
            self.__buffer = QBuffer()
            self.__stream = QDataStream()

        @property
        def data(self) -> QByteArray:
            return self.__data

        @data.setter
        def data(self, value):
            self.__data = value

        @property
        def version(self) -> int:
            return self.__version

        @version.setter
        def version(self, value):
            self.__version = value

        @property
        def buffer(self) -> QBuffer:
            return self.__buffer

        @property
        def stream(self) -> QDataStream:
            return self.__stream

    class EncryptedDescriptor(BaseObject):
        def __init__(self, size: int = None) -> None:
            self.__data = QByteArray()
            self.__buffer = QBuffer()
            self.__stream = QDataStream()

            if size:
                fullSize = 4 + size
                if fullSize & 0x0F:
                    fullSize += 0x10 - (fullSize & 0x0F)

                self.__data.reserve(fullSize)
                self.__data.resize(4)
                self.__buffer.setBuffer(self.__data)
                self.__buffer.open(QIODevice.OpenModeFlag.WriteOnly)
                self.__buffer.seek(4)
                self.__stream.setDevice(self.__buffer)
                self.__stream.setVersion(QDataStream.Version.Qt_5_1)

        def finish(self) -> None:
            if self.__stream.device():
                self.__stream.setDevice(None)  # type: ignore
            if self.__buffer.isOpen():
                self.__buffer.close()
            self.__buffer.setBuffer(None)  # type: ignore

        @property
        def data(self) -> QByteArray:
            return self.__data

        @data.setter
        def data(self, value):
            self.__data = value

        @property
        def buffer(self) -> QBuffer:
            return self.__buffer

        @property
        def stream(self) -> QDataStream:
            return self.__stream

    class FileWriteDescriptor(BaseObject):
        def __init__(self, fileName: str, basePath: str, sync: bool = False) -> None:
            self.fileName = fileName
            self.basePath = basePath
            self.sync = sync
            self.init(fileName)

        def init(self, fileName: str):
            self.base = Storage.PathJoin(self.basePath, fileName)
            self.safeData = QByteArray()
            self.buffer = QBuffer()
            self.buffer.setBuffer(self.safeData)

            result = self.buffer.open(QIODevice.OpenModeFlag.WriteOnly)
            Expects(result, "Can not open buffer, something went wrong")

            self.stream = QDataStream()
            self.stream.setDevice(self.buffer)

            self.md5 = bytes()
            self.fullSize = 0

        def writeData(self, data: QByteArray):
            Expects(self.stream.device() != None, "stream.device is missing")

            self.stream << data
            len = 0xFFFFFFFF if data.isNull() else data.size()

            if QSysInfo.Endian.ByteOrder != QSysInfo.Endian.BigEndian:

                def qbswap(source: int) -> int:
                    return (
                        0
                        | ((source & 0x000000FF) << 24)
                        | ((source & 0x0000FF00) << 8)
                        | ((source & 0x00FF0000) >> 8)
                        | ((source & 0xFF000000) >> 24)
                    )

                len = qbswap(len)

            self.md5 += len.to_bytes(4, "little")
            self.md5 += data.data()
            self.fullSize += 4 + data.size()

        def writeEncrypted(self, data: Storage.EncryptedDescriptor, key: td.AuthKey):
            self.writeData(Storage.PrepareEncrypted(data, key))

        def finish(self):
            if not self.stream.device():
                return
            self.stream.setDevice(None)  # type: ignore

            self.md5 += self.fullSize.to_bytes(4, "little")
            self.md5 += APP_VERSION.to_bytes(4, "little")
            self.md5 += TDF_MAGIC
            self.md5 = hashlib.md5(self.md5).digest()

            self.buffer.close()

            finalData: bytes = self.safeData.data() + self.md5
            Storage.WriteFile(self.fileName, self.basePath, finalData)

    @staticmethod
    def PrepareEncrypted(data: EncryptedDescriptor, key: td.AuthKey) -> QByteArray:
        data.finish()

        toEncrypt = QByteArray(data.data)

        # prepare for encryption
        size = toEncrypt.size()
        fullSize = size
        if fullSize & 0x0F:
            fullSize += 0x10 - (fullSize & 0x0F)
            toEncrypt.resize(fullSize)
            # TO BE ADDED
            # base::RandomFill(toEncrypt.data() + size, fullSize - size);

        # *(uint32*)toEncrypt.data() = size;
        toEncrypt = QByteArray(size.to_bytes(4, "little")) + toEncrypt[4:]

        hashData = hashlib.sha1(toEncrypt)
        encrypted = QByteArray(hashData.digest())

        encrypted.resize(0x10 + fullSize)  # 128bit of sha1 - key128, sizeof(data), data

        encrypted = encrypted[:0x10] + Storage.aesEncryptLocal(
            toEncrypt, key, encrypted
        )

        return encrypted

    @staticmethod
    def WriteFile(fileName: str, basePath: str, data: bytes):

        dir = QDir(basePath)
        if not dir.exists():
            dir.mkpath(basePath)

        file = QFile(Storage.PathJoin(basePath, fileName + "s"))
        if file.open(QIODevice.OpenModeFlag.WriteOnly):
            file.write(TDF_MAGIC)
            file.write(APP_VERSION.to_bytes(4, "little"))
            file.write(data)
            file.close()
            return True

        raise OpenTeleException("what")

    @staticmethod
    def ReadFile(fileName: str, basePath: str) -> FileReadDescriptor:

        to_try = ["s", "1", "0"]
        tries_exception = None
        for chr in to_try:
            try:

                file = QFile(Storage.PathJoin(basePath, fileName + chr))
                if not file.open(QIODevice.OpenModeFlag.ReadOnly):
                    tries_exception = TFileNotFound(f"Could not open {fileName}")
                    continue

                magic = file.read(4)

                if magic != TDF_MAGIC:
                    tries_exception = TDataInvalidMagic(
                        "Invalid magic {magic} in file {fileName}"
                    )
                    file.close()
                    continue

                version = int.from_bytes(file.read(4), "little")

                bytesdata = QByteArray(file.read(file.size()))

                dataSize = bytesdata.size() - 16

                check_md5 = bytesdata.data()[:dataSize]
                check_md5 += int(dataSize).to_bytes(4, "little")
                check_md5 += int(version).to_bytes(4, "little")
                check_md5 += magic
                check_md5 = hashlib.md5(check_md5)

                md5 = bytesdata.data()[dataSize:]

                if check_md5.hexdigest() != md5.hex():
                    tries_exception = TDataInvalidCheckSum(
                        "Invalid checksum {check_md5.hexdigest()} in file {fileName}"
                    )
                    file.close()
                    continue

                result = Storage.FileReadDescriptor()
                result.version = version

                bytesdata.resize(dataSize)
                result.data = bytesdata
                bytesdata = QByteArray()

                result.buffer.setBuffer(result.data)

                result.buffer.open(QIODevice.OpenModeFlag.ReadOnly)
                result.stream.setDevice(result.buffer)
                result.stream.setVersion(QDataStream.Version.Qt_5_1)

                file.close()
                return result
            except IOError as e:
                pass

        raise tries_exception if tries_exception else TFileNotFound(
            f"Could not open {fileName}"
        )

    @staticmethod
    def ReadEncryptedFile(
        fileName: str, basePath: str, authKey: td.AuthKey
    ) -> FileReadDescriptor:

        result = Storage.ReadFile(fileName, basePath)
        encrypted = QByteArray()
        result.stream >> encrypted

        try:
            data = Storage.DecryptLocal(encrypted, authKey)
        except OpenTeleException as e:
            result.stream.setDevice(None)  # type: ignore
            if result.buffer.isOpen():
                result.buffer.close()
            result.buffer.setBuffer(None)  # type: ignore
            result.data = QByteArray()
            result.version = 0
            raise e

        result.stream.setDevice(None)  # type: ignore
        if result.buffer.isOpen():
            result.buffer.close()
        result.buffer.setBuffer(None)  # type: ignore
        result.data = data.data

        result.buffer.setBuffer(result.data)
        result.buffer.open(QIODevice.OpenModeFlag.ReadOnly)
        result.buffer.seek(data.buffer.pos())
        result.stream.setDevice(result.buffer)
        result.stream.setVersion(QDataStream.Version.Qt_5_1)

        return result

    @staticmethod
    def ReadSetting(
        blockId: int, stream: QDataStream, version: int, context: ReadSettingsContext
    ) -> bool:  # pragma: no cover

        if blockId == dbi.DcOptionOldOld:
            dcId = DcId(stream.readUInt32())
            host = stream.readQString()
            ip = stream.readQString()
            port = stream.readUInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyDcOptions.constructAddOne(
                dcId, td.MTP.DcOptions.Flag(0), ip, port, bytes()
            )
            context.legacyRead = True

        elif blockId == dbi.DcOptionOld:
            dcIdWithShift = ShiftedDcId(stream.readUInt32())
            flags = td.MTP.DcOptions.Flag(stream.readInt32())
            ip = stream.readQString()
            port = stream.readUInt32()

            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyDcOptions.constructAddOne(
                dcIdWithShift, flags, ip, port, bytes()
            )
            context.legacyRead = True

        elif blockId == dbi.DcOptionsOld:
            serialized = QByteArray()
            stream >> serialized
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyDcOptions.constructFromSerialized(serialized)
            context.legacyRead = True

        elif blockId == dbi.ApplicationSettings:
            serialized = QByteArray()
            stream >> serialized
            ExpectStreamStatus(stream)

            # TO BE ADDED

        elif blockId == dbi.ChatSizeMaxOld:
            maxSize = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyChatSizeMax = maxSize
            context.legacyRead = True

        elif blockId == dbi.SavedGifsLimitOld:
            limit = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacySavedGifsLimit = limit
            context.legacyRead = True

        elif blockId == dbi.StickersRecentLimitOld:
            limit = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyStickersRecentLimit = limit
            context.legacyRead = True

        elif blockId == dbi.StickersFavedLimitOld:
            limit = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyStickersFavedLimit = limit
            context.legacyRead = True

        elif blockId == dbi.MegagroupSizeMaxOld:
            maxSize = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyMegagroupSizeMax = maxSize
            context.legacyRead = True

        elif blockId == dbi.User:
            userId = stream.readInt32()
            dcId = stream.readUInt32()
            ExpectStreamStatus(stream)

            context.mtpLegacyMainDcId = dcId
            context.mtpLegacyUserId = userId

        elif blockId == dbi.Key:
            dcId = DcId(stream.readInt32())
            key = stream.readRawData(256)
            ExpectStreamStatus(stream)

            context.mtpLegacyKeys.append(
                td.AuthKey(key, td.AuthKeyType.ReadFromFile, dcId)
            )

        elif blockId == dbi.MtpAuthorization:
            serialized = QByteArray()
            stream >> serialized
            ExpectStreamStatus(stream)

            context.mtpAuthorization = serialized

        return True

    @staticmethod
    def CreateLocalKey(
        salt: QByteArray, passcode: QByteArray = QByteArray()
    ) -> td.AuthKey:
        hashKey = hashlib.sha512(salt)
        hashKey.update(passcode)
        hashKey.update(salt)

        iterationsCount = 1 if passcode.isEmpty() else 100000
        return td.AuthKey(
            hashlib.pbkdf2_hmac("sha512", hashKey.digest(), salt, iterationsCount, 256)
        )

    @staticmethod
    def CreateLegacyLocalKey(
        salt: QByteArray, passcode: QByteArray = QByteArray()
    ) -> td.AuthKey:

        iterationsCount = 1 if passcode.isEmpty() else 100000
        return td.AuthKey(
            hashlib.pbkdf2_hmac(
                "sha512", passcode.data(), salt.data(), iterationsCount, 256
            )
        )

    @staticmethod
    def aesEncryptLocal(
        src: QByteArray, authKey: td.AuthKey, key128: QByteArray
    ) -> QByteArray:
        aesKey, aesIv = authKey.prepareAES_oldmtp(key128, False)
        encrypted = tgcrypto.ige256_encrypt(src, aesKey, aesIv)
        return QByteArray(encrypted)

    @staticmethod
    def aesDecryptLocal(
        src: QByteArray, authKey: td.AuthKey, key128: QByteArray
    ) -> QByteArray:
        aesKey, aesIv = authKey.prepareAES_oldmtp(key128, False)
        decrypted = tgcrypto.ige256_decrypt(src, aesKey, aesIv)
        return QByteArray(decrypted)

    @staticmethod
    def DecryptLocal(encrypted: QByteArray, authKey: td.AuthKey) -> EncryptedDescriptor:

        encryptedSize = encrypted.size()
        if (encryptedSize <= 16) or (encryptedSize & 0x0F):
            raise TDataBadEncryptedDataSize("Bad encrypted part size: {encryptedSize}")

        fullLen = encryptedSize - 16
        encryptedKey = encrypted[:16]
        encryptedData = encrypted[16:]

        decrypted = Storage.aesDecryptLocal(encryptedData, authKey, encryptedKey)
        checkHash = hashlib.sha1(decrypted).digest()[:16]
        if checkHash != encryptedKey:
            raise TDataBadDecryptKey(
                "Bad decrypt key, data not decrypted - incorrect password?"
            )

        dataLen = int.from_bytes(
            decrypted[:4], "little"
        )  # *(const uint32*)decrypted.constData();
        if (dataLen > decrypted.size()) or (dataLen <= fullLen - 16) or (dataLen < 4):
            raise TDataBadDecryptedDataSize(
                "Bad decrypted part size: {encryptedSize}, fullLen: {fullLen}, decrypted size: {decrypted.__len__()}"
            )

        decrypted.resize(dataLen)
        result = Storage.EncryptedDescriptor()
        result.data = decrypted

        decrypted = QByteArray()

        result.buffer.setBuffer(result.data)
        result.buffer.open(QIODevice.OpenModeFlag.ReadOnly)
        result.buffer.seek(4)  # skip len
        result.stream.setDevice(result.buffer)
        result.stream.setVersion(QDataStream.Version.Qt_5_1)

        return result

    @staticmethod
    def ComposeDataString(dataName: str, index: int):
        result = dataName
        result.replace("#", "")
        if index > 0:
            result += f"#{index + 1}"
        return result

    @staticmethod
    def ComputeDataNameKey(dataName: str) -> int:
        md5 = hashlib.md5(dataName.encode("utf-8")).digest()
        return int.from_bytes(md5, "little")

    @staticmethod
    def ToFilePart(val: int):
        result = str()
        for i in range(0, 0x10):
            v = val & 0xF
            if v < 0x0A:
                result += chr(ord("0") + v)
            else:
                result += chr(ord("A") + (v - 0x0A))
            val >>= 4
        return result

    @staticmethod
    def RandomGenerate(size: int) -> QByteArray:
        return QByteArray(os.urandom(size))

    @staticmethod
    def GetAbsolutePath(path: str = None) -> str:

        if path == None or path == "":
            path = os.getcwd()

        path = os.path.abspath(path)
        return path

    @staticmethod
    def PathJoin(path: str, filename: str) -> str:
        return os.path.join(path, filename)
